import os
from datetime import date, time, datetime, timedelta
from typing import List, Dict, Set, Union

# import from package
from resistics.calculators.calculator import Calculator
from resistics.calculators.decimationParameters import DecimationParams
from resistics.calculators.windowParameters import WindowParams
from resistics.dataObjects.projectData import ProjectData
from resistics.dataObjects.siteData import SiteData
from resistics.dataObjects.maskData import MaskData
from resistics.ioHandlers.spectrumReader import SpectrumReader
from resistics.ioHandlers.maskIO import MaskIO
from resistics.utilities.utilsPrint import list2ranges, blockPrint
from resistics.utilities.utilsWindow import datetime2gIndex, gIndex2datetime


class WindowSelector(Calculator):
    """Select windows for further processing

    Finds windows for further processing. Given more than one site, WindowSelector will find the shared windows. For example, when processing with a remote reference, WindowSelector will find the shared global windows (i.e. referenced to the project reference time) of the local site and the remote site.

    Shared windows will accept maskData when constraints are to be used. Further, datetime constraints can be added to the selection process, in the case where, for example, only night time data is to be processed.

    For datetime constraints, the order of priorities are:
    1. datetime constraints
    2. date constraints
    3. time constraints 


    Attributes
    ----------
    projData : ProjectData
        A ProjectData instance
    sampleFreq : float
        Sampling frequency of the data
    decParams : DecimationParameters
        A DecimationParameters instance detailing the decimation schemes
    winParams : WindowParameters
        A WindowParameters instance detailing the windowing information 
    sites : List[str]
        List of sites for which to calculate shared windows
    sharedWindows : Dict
        Dictionary to store shared window information between sites. The keys of the dictionary are the decimation levels.
    siteMasks : Dict
        Masks to use for each site
    siteSpecFolders : Dict
        Spectra folders for each site at sampleFreq
    siteSpecReaders : Dict
        SpectraReaders for each site at sampleFreq
    siteSpecRanges : Dict
        Global window ranges for all the spectra files for the sites at sampleFreq
    siteGlobalIndices : Dict
        Global window sets for all the spectra files for the sites at sampleFreq
    specdir : str
        The spectra data to use for calculating shared windows and the subsequent processing
    prepend : str
        The string prepending the spectra data files. This is usually "spectra" and not expected to change
    datetimes : Dict
        User supplied datetime constraints. These are the highest priority in the time constraints. This is a list of constraints for each decimation level.
    dates : Dict
        User supplied date constraints. These are the second highest priority in the time constraints. This is a list of constraints for each decimation level.
    times : Dict
        User supplied time constraints. These are the last priority in the time constraints. This is a list of constraints for each decimation level.
    datetimeConstraints : Dict
        Final combined date and time constraints. This is a list of constraints for each decimation level

    Methods
    -------
    __init__(decParams)
        Initialise window selector with information about the decimation parameters
    getSharedWindowsLevel(declevel)
        Get the shared windows for a decimation level as a python set
    getNumSharedWindows(declevel)
        Get the number of shared windows for a decimation level
    getWindowsForFreq(declevel, eIdx)
        Get the number of shared windows for a decimation level and evaluation frequency
    getUnmaskedWindowsLevel(declevel)
        Get unmasked windows for a decimation level
    getDatetimeConstraints()
        Get the datetime constraints
    getLevelDatetimeConstraints(declevel)
        Get the datetime constraints for a decimation level
    getMasks()
        Get a dictionary with masks to use for each site in the window selector
    getSpecReaderForWindow(site, declevel, iWin)
        Get the spectrum reader for a window
    getDataSize(declevel)
        Get the spectrum reader for a window
    setSites(sites)
        Set the sites for which to find the shared windows
    addDatetimeConstraint(start, stop)
        Add a datetime constraint
    addDateConstraint(dateC)
        Add a date constraint
    addTimeConstraint(start, stop)
        Add a time constraint. This will recur on every day of recording
    addWindowMask(site, maskName)
        Add a window mask
    resetDatetimeConstraints()
        Clear and reset all datetime constraints
    resetMasks()
        Clear and reset site masks        
    calcSharedWindows()
        Calculate shared windows between sites
    calcGlobalIndices()
        Find all the global indices for the sites
    calcDatetimeConstraints()
        Calculate overall datetime constraints
    calcSiteDates()
        Calculate a list of days that all the sites were operating
    printList()
        Class information as a list of strings
    printSiteInfo()
        Print out information about the sites included in the window selection
    printSiteInfoList(site)
        Return site window information as a list of strings
    printSharedWindows() 
        Print out the shared windows
    printSharedWindowsList()
        Shared window information as a list of strings
    printDatetimeConstraints()
        Print out the datetime constraints
    printDatetimeConstraintsList()
        Datetime constraint information as a list of strings
    printWindowMasks()
        Print information about masks being used in the window selection
    printWindowMasksList()
        Window mask information as a list of strings
    printWindowsForFrequency(listwindows=False):
        Print information about the windows for each evaluation frequency
    printWindowsForFrequencyList(listwindows=False)
        Information about windows for each evaluation frequency as a list of strings
    """

    def __init__(
        self,
        projData: ProjectData,
        sampleFreq: float,
        decParams: DecimationParams,
        winParams: WindowParams,
        **kwargs
    ) -> None:
        """Initialise window selector

        Parameters
        ----------
        projData : ProjectData
            A ProjectData instance
        sampleFreq : float
            The sampling frequency of the raw time data
        decParams : DecimationParams
            A decimation parameters instance detailing decimaion scheme
        winParams : WindowParams
            A window parameters instance detailing window schemes
        specdir : str, optional
            The spectra directories to use 
        """

        self.projData: ProjectData = projData
        self.sampleFreq: float = sampleFreq
        self.decParams = decParams
        self.winParams = winParams
        self.sites: List = []
        # shared indices
        self.sharedWindows: Dict = {}
        # the masks to use for each site - there can be multiple masks for each site
        self.siteMasks: Dict[str, List[str]] = {}
        # the spec files for each site at fs
        self.siteSpecFolders: Dict = {}
        self.siteSpecReaders: Dict = {}
        # global indices (ranges and sets)
        self.siteSpecRanges: Dict = {}
        self.siteGlobalIndices: Dict = {}
        # spectra directory information
        self.specdir = kwargs["specdir"] if "specdir" in kwargs else "spectra"
        self.prepend: str = "spectra"
        # time constraints: priority is datetimes > dates > times
        self.datetimes: Dict[int, List] = {}
        self.dates: Dict[int, List] = {}
        self.times: Dict[int, List] = {}
        # dictionary for datetime constraints
        self.datetimeConstraints: Dict[int, List] = {}
        # set all datetime constraints to empty
        self.resetDatetimeConstraints()

    def getSharedWindowsLevel(self, declevel: int) -> Set:
        """Get the shared windows for a decimation level

        Parameters
        ----------
        declevel : int
            The decimation level (0 is the first level)
        
        Returns
        -------
        set
            The shared windows for the decimation level
        """

        return self.sharedWindows[declevel]

    def getNumSharedWindows(self, declevel: int) -> int:
        """Get the number of shared windows for a decimation level

        Parameters
        ----------
        declevel : int
            The decimation level (0 is the first level)
        
        Returns
        -------
        int
            The number of shared windows for the decimation level
        """

        return len(self.sharedWindows[declevel])

    def getWindowsForFreq(self, declevel: int, eIdx: int) -> Set:
        """Get the number of shared windows for a decimation level and evaluation frequency

        Parameters
        ----------
        declevel : int
            The decimation level (0 is the first level)
        eIdx : int
            The evaluation frequency index
        
        Returns
        -------
        set
            The shared windows for evaluation frequency eIdx at decimation level declevel
        """

        sharedWindows = self.getSharedWindowsLevel(declevel)
        # now mask for the particular frequency - mask for each given site
        for s in self.sites:
            for mask in self.siteMasks[s]:
                # remove the masked windows from shared indices
                sharedWindows = sharedWindows - mask.getMaskWindowsFreq(declevel, eIdx)
        return sharedWindows

    def getUnmaskedWindowsLevel(self, declevel: int) -> Set:
        """Get unmasked windows for a decimation level

        Calculate the number of non masked windows for the decimation level. This should speed up processing when constraints are applied.

        Parameters
        ----------
        declevel : int
            The decimation level

        Returns
        -------
        set
            Unmasked windows for the decimation level
        """

        indices = set()
        evalFreq = self.decParams.getEvalFrequenciesForLevel(declevel)
        for eIdx, eFreq in enumerate(evalFreq):
            indices.update(self.getWindowsForFreq(declevel, eIdx))
        return indices

    def getDatetimeConstraints(self) -> Dict:
        """Get the datetime constraints
        
        Returns
        -------
        Dict
            Dictionary of datetime constraints at all decimation levels
        """

        self.calcDatetimeConstraints()
        return self.datetimeConstraints

    def getLevelDatetimeConstraints(self, declevel: int) -> List[List[datetime]]:
        """Get the datetime constraints for a decimation level
    
        Returns
        -------
        List[List[datetime]]
            Returns a list of datetime constraints, where each is a 2 element list with a start and stop
        """

        self.calcDatetimeConstraints()
        return self.datetimeConstraints[declevel]

    def getMasks(self) -> Dict:
        """Get a dictionary with masks to use for each site in the window selector
        
        Returns
        -------
        Dict
            Dictionary with masks to use for each site in the window selector
        """

        return self.siteMasks

    def getSpecReaderForWindow(self, site: str, declevel: int, iWin: int):
        """Get the spectrum reader for a window

        Parameters
        ----------
        site : str
            The name of the site to get the spectrum reader for
        declevel : int
            The decimation level
        iWin : int
            The window index
        
        Returns 
        -------
        specFile : str, bool
            The name of the spectra file or False if the window is not found in any spectra file
        specReader : SpectrumReader, bool
            The spectrum reader or False if the window is not found in any spectra file
        """

        specRanges = self.siteSpecRanges[site][declevel]
        specReaders = self.siteSpecReaders[site][declevel]
        for specFile in specRanges:
            if iWin >= specRanges[specFile][0] and iWin <= specRanges[specFile][1]:
                return specFile, specReaders[specFile]

        # if here, no window found
        self.printWarning(
            "Shared window {}, decimation level {} does not appear in any files given the constraints applied".format(
                iWin, declevel
            )
        )
        return False, False

    def getDataSize(self, declevel: int) -> int:
        """Get the spectrum reader for a window

        Parameters
        ----------
        declevel : str
            The decimation level

        Returns 
        -------
        int 
            The data size (number of points in the spectrum) at the decimation level
        """

        # return data size of first file
        site = self.sites[0]
        specReaders = self.siteSpecReaders[site][declevel]
        for sF in specReaders:
            return specReaders[sF].getDataSize()

    def setSites(self, sites: List[str]) -> None:
        """Set the sites for which to find the shared windows

        Parameters
        ----------
        sites : List[str]
            List of sites
        """

        # first remove repeated sites
        sitesSet = set(sites)
        sites = list(sitesSet)
        # now continue
        self.sites = sites
        for s in self.sites:
            self.siteMasks[s] = []
            self.siteSpecFolders[s] = []
            self.siteSpecReaders[s] = {}
            self.siteSpecRanges[s] = {}
            # use sets to hold gIndices
            # optimised to find intersections
            self.siteGlobalIndices[s] = {}
        # at the end, calculate global indices
        self.calcGlobalIndices()

    def addDatetimeConstraint(
        self, start: str, stop: str, declevel: Union[List[int], int, None] = None
    ):
        """Add datetime constraints

        Parameters
        ----------
        start : str
            Datetime constraint start in format %Y-%m-%d %H:%M:%S
        stop : str
            Datetime constraint end in format %Y-%m-%d %H:%M:%S
        declevel : List[int], int, optional
            The decimation level. If left as default, will be applied to all decimation levels.             
        """

        datetimeStart = datetime.strptime(start, "%Y-%m-%d %H:%M:%S")
        datetimeStop = datetime.strptime(stop, "%Y-%m-%d %H:%M:%S")

        # levels the constraint applies to
        if declevel is None:
            levels = range(0, self.decParams.numLevels)
        elif isinstance(declevel, list):
            levels = declevel
        else:
            levels = [declevel]
        # then add constraints as appropriate
        for declevel in levels:
            self.datetimes[declevel].append([datetimeStart, datetimeStop])

    def addDateConstraint(
        self, dateC: str, declevel: Union[List[int], int, None] = None
    ):
        """Add a date constraint

        Parameters
        ----------
        dateC : str
            Datetime constraint in format %Y-%m-%d
        declevel : List[int], int, optional
            The decimation level. If left as default, will be applied to all decimation levels.             
        """

        datetimeC = datetime.strptime(dateC, "%Y-%m-%d").date()

        # levels the constraint applies to
        if declevel is None:
            levels = range(0, self.decParams.numLevels)
        elif isinstance(declevel, list):
            levels = declevel
        else:
            levels = [declevel]
        # then add constraints as appropriate
        for declevel in levels:
            self.dates[declevel].append(datetimeC)

    def addTimeConstraint(
        self, start: str, stop: str, declevel: Union[List[int], int, None] = None
    ):
        """Add a time constraint. This will recur on every day of recording.

        Parameters
        ----------
        start : str
            Time constraint start in format %H:%M:%S
        stop : str
            Time constraint end in format %H:%M:%S
        declevel : List[int], int, optional
            The decimation level. If left as default, will be applied to all decimation levels.             
        """

        timeStart = datetime.strptime(start, "%H:%M:%S").time()
        timeStop = datetime.strptime(stop, "%H:%M:%S").time()

        # levels the constraint applies to
        if declevel is None:
            levels = range(0, self.decParams.numLevels)
        elif isinstance(declevel, list):
            levels = declevel
        else:
            levels = [declevel]
        # then add constraints as appropriate
        for declevel in levels:
            self.times[declevel].append([timeStart, timeStop])

    def addWindowMask(self, site: str, maskName: str) -> None:
        """Add a window mask
        
        This is a mask with values for each evaluation frequency.

        Parameters
        ----------
        site : str
            The site for which to search for a mask
        maskName : str
            The name of the mask
        """

        siteData = self.projData.getSiteData(site)
        maskIO = MaskIO(siteData.getSpecdirMaskPath(self.specdir))
        maskData = maskIO.read(maskName, self.sampleFreq)
        self.siteMasks[site].append(maskData)

    def resetDatetimeConstraints(self) -> None:
        """Reset datetime constraints"""

        # add a list for each decimation level
        for declevel in range(0, self.decParams.numLevels):
            self.datetimes[declevel] = []
            self.dates[declevel] = []
            self.times[declevel] = []
            self.datetimeConstraints[declevel] = []

    def resetMasks(self) -> None:
        """Reset masks"""

        # reset to no masks for any site
        for site in self.siteMasks:
            self.siteMasks[site] = []

    def calcSharedWindows(self):
        """Calculate shared windows between sites

        Calculates the shared windows between sites. Datetime constraints are applied. No masks are applied in this method. Masks are only applied when getting the windows for a particular evaluation frequency.
        """

        if len(self.sites) == 0:
            self.printWarning(
                "No sites given to Window Selector. At least one site needs to be given."
            )
            return False

        # calculate datetime constraints
        self.calcDatetimeConstraints()

        # initialise the sharedWindows with a set from one site
        sites = self.sites
        siteInit = sites[0]
        numLevels = self.decParams.numLevels
        for declevel in range(0, numLevels):
            self.sharedWindows[declevel] = self.siteGlobalIndices[siteInit][declevel]

        # now for each decimation level, calculate the shared windows
        for declevel in range(0, numLevels):
            for site in self.sites:
                self.sharedWindows[declevel] = self.sharedWindows[
                    declevel
                ].intersection(self.siteGlobalIndices[site][declevel])

        # apply time constraints
        # time constraints should be formulated as a set
        # and then, find the intersection again
        for declevel in range(0, numLevels):
            constraints = self.getLevelDatetimeConstraints(declevel)
            if len(constraints) != 0:
                datetimeIndices = set()
                for dC in constraints:
                    gIndexStart, firstWindowStart = datetime2gIndex(
                        self.projData.refTime,
                        dC[0],
                        self.decParams.getSampleFreqLevel(declevel),
                        self.winParams.getWindowSize(declevel),
                        self.winParams.getOverlap(declevel),
                    )
                    gIndexEnd, firstWindowEnd = datetime2gIndex(
                        self.projData.refTime,
                        dC[1],
                        self.decParams.getSampleFreqLevel(declevel),
                        self.winParams.getWindowSize(declevel),
                        self.winParams.getOverlap(declevel),
                    )
                    gIndexEnd = (
                        gIndexEnd - 1
                    )  # as the function returns the next window starting after time
                    if gIndexEnd < gIndexStart:
                        gIndexEnd = gIndexStart
                    datetimeIndices.update(list(range(gIndexStart, gIndexEnd)))
                    self.printText(
                        "Decimation level = {}. Applying date constraint {} - {}, global index constraint {} - {}".format(
                            declevel, dC[0], dC[1], gIndexStart, gIndexEnd
                        )
                    )
                self.sharedWindows[declevel] = self.sharedWindows[
                    declevel
                ].intersection(datetimeIndices)

    def calcGlobalIndices(self) -> None:
        """Find all the global indices for the sites"""

        # get all the spectra files with the correct sampling frequency
        for site in self.sites:
            siteData = self.projData.getSiteData(site)
            timeFilesFs = siteData.getMeasurements(self.sampleFreq)
            # specFiles = self.proj.getSiteSpectraFiles(s)
            specFolders = siteData.spectra
            specFoldersFs = []
            for specFolder in specFolders:
                if specFolder in timeFilesFs:
                    specFoldersFs.append(specFolder)

            self.siteSpecFolders[site] = specFoldersFs

            # for each decimation level
            # loop through each of the spectra folders
            # and find the global indices ranges for each decimation level
            numLevels = self.decParams.numLevels
            for declevel in range(0, numLevels):
                # get the dictionaries ready
                self.siteSpecReaders[site][declevel] = {}
                self.siteSpecRanges[site][declevel] = {}
                self.siteGlobalIndices[site][declevel] = set()
                # loop through spectra folders and figure out global indices
                for specFolder in self.siteSpecFolders[site]:
                    # here, have to use the specdir option
                    specReader = SpectrumReader(
                        os.path.join(siteData.specPath, specFolder, self.specdir)
                    )
                    # here, use prepend to open the spectra file
                    check = specReader.openBinaryForReading(self.prepend, declevel)
                    # if file does not exist, continue
                    if not check:
                        continue
                    self.siteSpecReaders[site][declevel][specFolder] = specReader
                    globalRange = specReader.getGlobalRange()
                    self.siteSpecRanges[site][declevel][specFolder] = globalRange
                    # and save set of global indices
                    self.siteGlobalIndices[site][declevel].update(
                        list(range(globalRange[0], globalRange[1] + 1))
                    )

    def calcDatetimeConstraints(self) -> None:
        """Calculate overall datetime constraints

        Priority order for datetime constraints is: 
        1. datetime constraints
        2. date constraints
        3. time constraints   
        """

        # calculate site dates if required
        siteDates = self.calcSiteDates()

        # datetime constraints are for each decimation level
        numLevels = self.decParams.numLevels
        for declevel in range(0, numLevels):
            # calculate date and time constraints for each level
            # begin with the datetime constraints - these have highest priority
            self.datetimeConstraints[declevel] = self.datetimes[declevel]

            # check to see whether any date and time constraints
            if len(self.dates[declevel]) == 0 and len(self.times[declevel]) == 0:
                continue

            dateConstraints = []
            if len(self.dates[declevel]) != 0:
                # apply time constraints only on specified days
                dateConstraints = self.dates[declevel]
            else:
                dateConstraints = siteDates

            # finally, add the time constraints to the dates
            # otherwise add the whole day
            dateAndTimeConstraints = []
            if len(self.times[declevel]) == 0:
                # add whole days
                for dC in dateConstraints:
                    start = datetime.combine(dC, time(0, 0, 0))
                    stop = datetime.combine(dC, time(23, 59, 59))
                    dateAndTimeConstraints.append([start, stop])
            else:
                # add each time for each day
                for tC in self.times[declevel]:
                    for dC in dateConstraints:
                        start = datetime.combine(dC, tC[0])
                        stop = datetime.combine(dC, tC[1])
                        # check if this goes over a day
                        if tC[1] < tC[0]:
                            # then need to increment the day
                            dCNext = dC + timedelta(days=1)
                            stop = datetime.combine(dCNext, tC[1])
                        # append to constraints
                        dateAndTimeConstraints.append([start, stop])

            # finally, combine datetimes and dateAndTimeConstraints
            self.datetimeConstraints[declevel] = (
                self.datetimeConstraints[declevel] + dateAndTimeConstraints
            )
            self.datetimeConstraints[declevel] = sorted(
                self.datetimeConstraints[declevel]
            )

    def calcSiteDates(self) -> List[datetime]:
        """Calculate a list of days that all the sites were operating
        
        This uses the siteStart and siteEnd datetimes, so does not take into account the start and end of actual time series measurements, which is taken into account later.

        Returns
        -------
        List[datetime]
            A list of dates all the sites were operating
        """

        starts = []
        stops = []
        for site in self.sites:
            siteData = self.projData.getSiteData(site)
            starts.append(siteData.siteStart)
            stops.append(siteData.siteEnd)
        # need all the dates between
        d1 = max(starts).date()
        d2 = min(stops).date()
        if d1 > d2:
            self.printError(
                "A site passed to the window selector does not overlap with any other sites. There will be no shared windows",
                quitRun=True,
            )
        # now with d2 > d1
        siteDates = []
        delta = d2 - d1
        # + 1 because inclusive of stop and start days
        for i in range(delta.days + 1):
            siteDates.append(d1 + timedelta(days=i))
        return siteDates

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : list
            List of strings with information
        """

        textLst = []
        textLst.append("Sampling frequency [Hz] = {:.6f}".format(self.sampleFreq))
        textLst.append("Spectra directory = {}".format(self.specdir))
        textLst.append("Sites = {}".format(", ".join(self.sites)))
        textLst.append("Site information:")
        for site in self.sites:
            textLst = textLst + self.printSiteInfoList(site)
        return textLst

    def printSiteInfo(self):
        """Print out information about the sites"""

        for site in self.sites:
            blockPrint("WindowSelector::site info", self.printSiteInfoList(site))

    def printSiteInfoList(self, site: str) -> List[str]:
        """Return site window information as a list of strings

        Parameters
        ----------
        site : str
            The site name
        
        Returns
        -------
        List[str]
            Site window information as a list of strings
        """

        textLst = []
        textLst.append("Sampling frequency [Hz] = {:.6f}".format(self.sampleFreq))
        textLst.append("Site = {}".format(site))
        textLst.append("Site global index information")
        numLevels = self.decParams.numLevels
        for declevel in range(0, numLevels):
            textLst.append("Decimation Level = {:d}".format(declevel))
            ranges = self.siteSpecRanges
            for sF in sorted(list(ranges[site][declevel].keys())):
                startTime1, endTime1 = gIndex2datetime(
                    ranges[site][declevel][sF][0],
                    self.projData.refTime,
                    self.sampleFreq / self.decParams.getDecFactor(declevel),
                    self.winParams.getWindowSize(declevel),
                    self.winParams.getOverlap(declevel),
                )
                startTime2, endTime2 = gIndex2datetime(
                    ranges[site][declevel][sF][1],
                    self.projData.refTime,
                    self.sampleFreq / self.decParams.getDecFactor(declevel),
                    self.winParams.getWindowSize(declevel),
                    self.winParams.getOverlap(declevel),
                )
                textLst.append(
                    "Measurement file = {}\ttime range = {} - {}\tGlobal Indices Range = {:d} - {:d}".format(
                        sF,
                        startTime1,
                        endTime2,
                        ranges[site][declevel][sF][0],
                        ranges[site][declevel][sF][1],
                    )
                )
        return textLst

    def printSharedWindows(self) -> None:
        """Print out the shared windows"""

        blockPrint("WindowSelector::shared windows", self.printSharedWindowsList())

    def printSharedWindowsList(self) -> List[str]:
        """Shared window information as a list of strings

        Returns
        -------
        List[str]
            Shared window information as a list of strings
        """

        textLst = []
        numLevels = self.decParams.numLevels
        for declevel in range(0, numLevels):
            textLst.append("Decimation Level = {:d}".format(declevel))
            textLst.append(
                "\tNumber of shared windows = {:d}".format(
                    self.getNumSharedWindows(declevel)
                )
            )
            textLst.append(
                "\tShared window indices: {}".format(
                    list2ranges(self.getSharedWindowsLevel(declevel))
                )
            )
            textLst.append(
                "\tNumber of unmasked windows: {}".format(
                    len(self.getUnmaskedWindowsLevel(declevel))
                )
            )
        textLst.append(
            "NOTE: These are the shared windows at each decimation level. Windows for each evaluation frequency might vary depending on masks"
        )
        return textLst

    def printDatetimeConstraints(self):
        """Print out the datetime constraints"""

        blockPrint(
            "WindowSelector::datetime constraints", self.printDatetimeConstraintsList()
        )

    def printDatetimeConstraintsList(self) -> List[str]:
        """Datetime constraint information as a list of strings

        Returns
        -------
        List[str]
            Datetime constraint information as a list of strings
        """

        textLst = []
        # calculate datetime constraints
        self.calcDatetimeConstraints()
        # populate textLst
        textLst.append("Datetime constraints")
        numLevels = self.decParams.numLevels
        for declevel in range(0, numLevels):
            textLst.append("Decimation Level = {:d}".format(declevel))
            for d in self.getLevelDatetimeConstraints(declevel):
                textLst.append("\tConstraint {} - {}".format(d[0], d[1]))
        return textLst

    def printWindowMasks(self) -> None:
        """Print mask information"""

        blockPrint("WindowSelector::window masks", self.printWindowMasksList())

    def printWindowMasksList(self) -> List[str]:
        """Window mask information as a list of strings

        Returns
        -------
        List[str]
            Window mask information as a list of strings
        """

        textLst = []
        for s in self.sites:
            textLst.append("Site = {}".format(s))
            if len(self.siteMasks[s]) == 0:
                textLst.append("\tNo masks for this site")
            else:
                for mask in self.siteMasks[s]:
                    textLst.append("\tMask = {}".format(mask.maskName))
        return textLst

    def printWindowsForFrequency(self, listwindows=False):
        """Print information about the windows for each evaluation frequency

        Parameters
        ----------
        listwindows : bool
            Boolean flag to actually write out all the windows. Default is False as this takes up a lot of space in the terminal
        """

        blockPrint(
            "WindowSelector::windows for frequency",
            self.printWindowsForFrequencyList(listwindows),
        )

    def printWindowsForFrequencyList(self, listwindows=False) -> List[str]:
        """Information about windows for each evaluation frequency as a list of strings

        Parameters
        ----------
        listwindows : bool
            Boolean flag to actually write out all the windows. Default is False as this takes up a lot of space in the terminal

        Returns
        -------
        List[str]
            Windows for evaluation frequency information as a list of strings
        """

        textLst = []
        for declevel in range(0, self.decParams.numLevels):
            evalFreq = self.decParams.getEvalFrequenciesForLevel(declevel)
            unmaskedWindows = self.getNumSharedWindows(declevel)
            for eIdx, eFreq in enumerate(evalFreq):
                maskedWindows = self.getWindowsForFreq(declevel, eIdx)
                textLst.append(
                    "Evaluation frequency = {:.6f}, shared windows = {:d}, windows after masking = {:d}".format(
                        eFreq, unmaskedWindows, len(maskedWindows)
                    )
                )
                if listwindows:
                    textLst.append("{}".format(list2ranges(maskedWindows)))
        return textLst



import numpy as np
import scipy.stats as stats
from typing import List, Dict, Tuple

# import from package
from resistics.utilities.utilsPrint import errorPrint


def getStatNames() -> Tuple[List[str], List[str]]:
    """Get a list of statistic and remotestatistic names

    Returns
    -------
    stats : List[str]
        List of signal site statistics
    remoteStats : List[str]
        List of remote reference statistics
    """

    stats = [
        "absvalEqn",
        "coherence",
        "powerSpectralDensity",
        "polarisationDirection",
        "partialCoherence",
        "transferFunction",
        "resPhase",
    ]
    remoteStats = [
        "RR_coherence",
        "RR_coherenceEqn",
        "RR_absvalEqn",
        "RR_transferFunction",
        "RR_resPhase",
    ]
    return stats, remoteStats


def getStatElements(stat: str) -> List[str]:
    """Get statistic elements for each statistic
    
    Parameters
    ----------
    stat : str
        The statistic for which to get the statistic elements

    Returns
    -------
    Dict[str, List[str]]
        Mapping from a statistic name to the elements of that statistic
    """

    statElements = {
        "absvalEqn": [
            "absExEx",
            "absHyEx",
            "absExEy",
            "absHyEy",
            "absExHx",
            "absHyHx",
            "absExHy",
            "absHyHy",
            "absEyEx",
            "absHxEx",
            "absEyEy",
            "absHxEy",
            "absEyHx",
            "absHxHx",
            "absEyHy",
            "absHxHy",
        ],
        "coherence": ["cohExHx", "cohExHy", "cohEyHx", "cohEyHy"],
        "powerSpectralDensity": ["psdEx", "psdEy", "psdHx", "psdHy"],
        "polarisationDirection": ["polExEy", "polHxHy"],
        "transferFunction": [
            "ExHxReal",
            "ExHxImag",
            "ExHyReal",
            "ExHyImag",
            "EyHxReal",
            "EyHxImag",
            "EyHyReal",
            "EyHyImag",
        ],
        "resPhase": [
            "ExHxRes",
            "ExHxPhase",
            "ExHyRes",
            "ExHyPhase",
            "EyHxRes",
            "EyHxPhase",
            "EyHyRes",
            "EyHyPhase",
        ],
        "partialCoherence": [
            "bivarEx",
            "bivarEy",
            "parExHx",
            "parExHy",
            "parEyHx",
            "parEyHy",
        ],
    }
    # remote reference stat elements
    statElementsRR = {
        "RR_coherence": [
            "ExHxRR",
            "ExHyRR",
            "EyHxRR",
            "EyHyRR",
            "HxHxRR",
            "HxHyRR",
            "HyHxRR",
            "HyHyRR",
        ],
        "RR_coherenceEqn": ["ExHxR-HyHxR", "ExHyR-HyHyR", "EyHxR-HxHxR", "EyHyR-HxHyR"],
        "RR_absvalEqn": [
            "absHyHxR",
            "absExHxR",
            "absHyHyR",
            "absExHyR",
            "absHxHxR",
            "absEyHxR",
            "absHxHyR",
            "absEyHyR",
        ],
        "RR_transferFunction": [
            "ExHxRealRR",
            "ExHxImagRR",
            "ExHyRealRR",
            "ExHyImagRR",
            "EyHxRealRR",
            "EyHxImagRR",
            "EyHyRealRR",
            "EyHyImagRR",
        ],
        "RR_resPhase": [
            "ExHxResRR",
            "ExHxPhaseRR",
            "ExHyResRR",
            "ExHyPhaseRR",
            "EyHxResRR",
            "EyHxPhaseRR",
            "EyHyResRR",
            "EyHyPhaseRR",
        ],
    }

    if stat in statElements:
        return statElements[stat]
    if stat in statElementsRR:
        return statElementsRR[stat]
    errorPrint(
        "utilsStats::getStatElements",
        "Statistic {} not found".format(stat),
        quitRun=True,
    )
    return False


# Window statistics - These are not currently in use
# signal to noise ratio is calculated as mean over std.
# this might be more useful in spectral domain
# can do this in the frequecy domain, on the amplitude
def calcSNR(specData):
    output = {}
    for c in specData.data:
        tmp = np.absolute(specData.data[c])
        output[c] = np.average(tmp) / np.std(tmp)
    return output


# The Pearson correlation coefficient measures the linear relationship
# between two datasets. Strictly speaking, Pearson's correlation requires
# that each dataset be normally distributed.
def pearsonCoefficient(data):
    # construct input matrices
    # this needs to be columns for observations, rows for variables
    # and this needs to be output as a dictionary
    chans = sorted(list(data.keys()))
    numChans = len(chans)  # the channels
    output = {}
    for i in range(0, numChans):
        for j in range(i, numChans):
            key = "{}{}".format(chans[i], chans[j])
            # now calculate the pearson correlation coefficient
            pcc, pval = stats.pearsonr(data[chans[i]], data[chans[j]])
            output[key] = pcc
    return output


def pearsonCoefficientSpec(data):
    # calcaulates the PCC for magnitude and phase separately
    mag = {}
    phase = {}
    for c in data:
        mag[c] = np.absolute(data[c])
        phase[c] = np.unwrap(np.angle(data[c]))
    magPCC = pearsonCoefficient(mag)
    phasePCC = pearsonCoefficient(phase)
    return magPCC, phasePCC



from datetime import datetime, timedelta
import calendar
import numpy as np
import math


def gIndex2datetime(
    gIndex: int, refTime: datetime, fs: float, windowSize: int, windowOverlap: int
):
    """Global index to datetime convertor

    Global index 0 corresponds to reference time

    Parameters
    ----------
    gIndex : int
        Globel index
    refTime : datetime.datetime
        Reference time   
    fs : float
        Sampling frequency in Hz
    windowSize : int
        Size of windows
    windowOverlap : int
        Size of window overlaps     

    Returns
    -------
    startTime : datetime.datetime
        Start time of global window gIndex
    endTime : datetime.datetime
        End time of global window gIndex        
    """

    # global index 0 starts at refTime
    timeOffset = 1.0 * (windowSize - windowOverlap) / fs
    totalOffset = gIndex * timeOffset
    startTime = refTime + timedelta(seconds=totalOffset)
    # windowSize - 1 because inclusive of start sample
    endTime = startTime + timedelta(seconds=1.0 * (windowSize - 1) / fs)
    return startTime, endTime


def gArray2datetime(
    gArray: np.ndarray,
    refTime: datetime,
    fs: float,
    windowSize: int,
    windowOverlap: int,
):
    """Global index array to datetime convertor

    Global index 0 corresponds to reference time

    Parameters
    ----------
    gArray : np.ndarray
        Globel indices array
    refTime : datetime.datetime
        Reference time   
    fs : float
        Sampling frequency in Hz
    windowSize : int
        Size of windows
    windowOverlap : int
        Size of window overlaps     

    Returns
    -------
    startTime : np.ndarray of datetime.datetime
        Start times of global windows
    endTime : np.ndarray of datetime.datetime
        End times of global windows         
    """

    arrSize = gArray.size
    startTime = np.zeros(shape=(arrSize), dtype=datetime)
    endTime = np.zeros(shape=(arrSize), dtype=datetime)
    for i in range(0, arrSize):
        startTime[i], endTime[i] = gIndex2datetime(
            gArray[i], refTime, fs, windowSize, windowOverlap
        )
    return startTime, endTime


def gIndex2timestamp(
    gIndex: int, refTime: datetime, fs: float, windowSize: int, windowOverlap: int
):
    """Global index to timestamp convertor

    Global index 0 corresponds to reference time

    Parameters
    ----------
    gIndex : int
        Globel index
    refTime : datetime.datetime
        Reference time   
    fs : float
        Sampling frequency in Hz
    windowSize : int
        Size of windows
    windowOverlap : int
        Size of window overlaps     

    Returns
    -------
    startTime : UNIX timestamp
        Start time of global window gIndex
    endTime : UNIX timestamp
        End time of global window gIndex        
    """

    startTime, endTime = gIndex2datetime(gIndex, refTime, fs, windowSize, windowOverlap)
    return calendar.timegm(startTime.timetuple()), calendar.timegm(endTime.timetuple())


def gArray2timestamp(
    gArray: np.ndarray,
    refTime: datetime,
    fs: float,
    windowSize: int,
    windowOverlap: int,
):
    """Global index array to timestamp convertor

    Global index 0 corresponds to reference time

    Parameters
    ----------
    gArray : np.ndarray
        Globel indices array
    refTime : datetime.datetime
        Reference time   
    fs : float
        Sampling frequency in Hz
    windowSize : int
        Size of windows
    windowOverlap : int
        Size of window overlaps     

    Returns
    -------
    startTime : UNIX timestamp
        Start times of global windows
    endTime : UNIX timestamp
        End times of global windows         
    """

    arrSize = gArray.size
    startTime = np.zeros(shape=(arrSize), dtype=datetime)
    endTime = np.zeros(shape=(arrSize), dtype=datetime)
    for i in range(0, arrSize):
        startTime[i], endTime[i] = gIndex2timestamp(
            gArray[i], refTime, fs, windowSize, windowOverlap
        )
    return startTime, endTime


def datetime2gIndex(
    refTime: datetime, inTime: datetime, fs: float, windowSize: int, windowOverlap: int
):
    """Datetime to global index convertor

    Global index 0 corresponds to reference time. This returns the global index of the time window nearest to inTime

    Parameters
    ----------
    refTime : datetime.datetime
        Reference time  
    inTime : datetime.datetime
        Time for which you want closest global index 
    fs : float
        Sampling frequency in Hz
    windowSize : int
        Size of windows
    windowOverlap : int
        Size of window overlaps     

    Returns
    -------
    gIndex : int
        Global window index closest to inTime
    firstWindowTime : datetime.datetime
        Datetime of the global window        
    """

    # need to return the next one close
    # calculate
    deltaRefStart = inTime - refTime
    winStartIncrement = (windowSize - windowOverlap) / fs
    # calculate number of windows started before reference time
    # and then by taking the ceiling, find the global index of the first window in the data
    gIndex = int(math.ceil(deltaRefStart.total_seconds() / winStartIncrement))
    # calculate start time of first global window
    offsetSeconds = gIndex * winStartIncrement
    # calculate the first window time
    firstWindowTime = refTime + timedelta(seconds=offsetSeconds)
    return gIndex, firstWindowTime



import os
from datetime import datetime
import glob
from typing import Union

# import from package
from resistics.dataObjects.projectData import ProjectData
from resistics.dataObjects.configData import ConfigData
from resistics.utilities.utilsConfig import loadConfig
from resistics.utilities.utilsIO import checkAndMakeDir
from resistics.utilities.utilsProject import (
    projectText,
    projectBlock,
    projectWarning,
    projectError,
)


def newProject(
    projectPath: str,
    refTime: Union[str, datetime],
    configFile: str = "",
    name: str = "mtProj",
) -> ProjectData:
    """Create a new project in project path

    A new project will be created in project path. If the project path directory does not exist, a new one will be made. If a project already exists in project path, this project will be loaded and returned.

    Parameters
    ----------
    projectPath : str
        Path for the project directory
    refTime : datetime
        The reference time for the project
    configFile : str, optional
        Path to a configuration file
    name : str, optional (default is "mtProj")
        The name of the project file

    Returns
    -------
    ProjectData
        A project data object
    """

    # check if a project file already exists and if so, load it

    # create project path directory
    checkAndMakeDir(projectPath)
    # reference time
    if isinstance(refTime, str):
        refTime = datetime.strptime(refTime, "%Y-%m-%d %H:%M:%S")
    # print info
    textLst = [
        "Creating a new project in path: {}".format(projectPath),
        "Project name: {}".format(name),
    ]
    projectBlock(textLst)

    # create the subdirectories
    projectFile = os.path.join(projectPath, "{}.prj".format(name))
    timePath = os.path.join(projectPath, "timeData")
    specPath = os.path.join(projectPath, "specData")
    statPath = os.path.join(projectPath, "statData")
    maskPath = os.path.join(projectPath, "maskData")
    transFuncPath = os.path.join(projectPath, "transFuncData")
    calPath = os.path.join(projectPath, "calData")
    imagePath = os.path.join(projectPath, "images")
    saveProjectFile(
        projectFile,
        refTime,
        calPath,
        timePath,
        specPath,
        statPath,
        maskPath,
        transFuncPath,
        imagePath,
    )

    # configuration file
    config = ConfigData(configFile)

    proj = ProjectData(
        projectFile,
        refTime,
        calPath,
        timePath,
        specPath,
        statPath,
        maskPath,
        transFuncPath,
        imagePath,
        config=config,
    )
    proj.printInfo()
    proj.config.printInfo()

    return proj


def loadProject(projectPath: str, configFile: str = "") -> ProjectData:
    """Load an existing project

    Parameters
    ----------
    projectPath : str
        Path for the project directory
    configFile : str
        Path to a configuration file

    Returns
    -------
    ProjectData
        A project data object
    """

    # search for the .prj file (hopefully only one)
    gl = glob.glob(os.path.join(projectPath, "*.prj"))
    if len(gl) == 0:
        projectError("Unable to find project file in path: {}".format(projectPath))
    projectFile: str = os.path.basename(gl[0])
    projectText(
        "Loading project file: {}".format(os.path.join(projectPath, projectFile))
    )
    projectPaths = loadProjectFile(os.path.join(projectPath, projectFile))

    # check the configuration file
    config = ConfigData(configFile)

    proj = ProjectData(
        projectFile,
        projectPaths["refTime"],
        projectPaths["calPath"],
        projectPaths["timePath"],
        projectPaths["specPath"],
        projectPaths["statPath"],
        projectPaths["maskPath"],
        projectPaths["transFuncPath"],
        projectPaths["imagePath"],
        config=config,
    )
    proj.printInfo()
    proj.config.printInfo()

    return proj


def saveProjectFile(
    filepath: str,
    refTime: datetime,
    calPath: str,
    timePath: str,
    specPath: str,
    statPath: str,
    maskPath: str,
    transFuncPath: str,
    imagePath: str,
) -> None:
    """Save the project file to filepath

    Parameters
    ----------
    filepath : str
        Path to project file
    """

    f = open(filepath, "w")
    f.write("Calibration data path = {}\n".format(calPath))
    f.write("Time data path = {}\n".format(timePath))
    f.write("Spectra data path = {}\n".format(specPath))
    f.write("Statistics data path = {}\n".format(statPath))
    f.write("Mask data path = {}\n".format(maskPath))
    f.write("TransFunc data path = {}\n".format(transFuncPath))
    f.write("Image data path = {}\n".format(imagePath))
    f.write("Reference time = {}\n".format(refTime.strftime("%Y-%m-%d %H:%M:%S")))
    f.close()


def loadProjectFile(filepath: str) -> None:
    """Load project from path to project file

    Parameters
    ----------
    filepath : str
        Path to project file

    Returns
    -------
    projPath : Dict[str, Union[str, datetime]]
        A dictionary with various project paths
    """

    f = open(filepath, "r")
    lines = f.readlines()
    f.close()
    projPaths = {}
    for l in lines:
        split = l.strip().split("=")
        if "Calibration data path" in l:
            projPaths["calPath"] = split[1].strip()
        if "Time data path" in l:
            projPaths["timePath"] = split[1].strip()
        if "Spectra data path" in l:
            projPaths["specPath"] = split[1].strip()
        if "Statistics data path" in l:
            projPaths["statPath"] = split[1].strip()
        if "Mask data path" in l:
            projPaths["maskPath"] = split[1].strip()
        if "TransFunc data path" in l:
            projPaths["transFuncPath"] = split[1].strip()
        if "Image data path" in l:
            projPaths["imagePath"] = split[1].strip()
        if "Reference time" in l:
            projPaths["refTime"] = datetime.strptime(
                split[1].strip(), "%Y-%m-%d %H:%M:%S"
            )
    return projPaths




import os
import numpy as np

# import from package
from resistics.ioHandlers.dataWriter import DataWriter


class DataWriterInternal(DataWriter):
    """Write out binary data files

    This is simply header files and binary data files. The header file saved is relevant only to this software and needs to be read in using DataReaderInternal.
    The header file means less processing to read the header information

    Methods
    -------
    writeDataFiles(chans, timeData)
        Write out time series data
    """

    def writeDataFiles(self, chans, timeData) -> None:
        for idx, c in enumerate(chans):
            writePath = os.path.join(
                self.getOutPath(), "chan_{:02d}{}".format(idx, self.extension)
            )
            dataF = open(writePath, "wb")
            dataF.write(timeData.data[c].astype(self.dtype).tobytes())
            dataF.close()



import os
import numpy as np
import math
from datetime import datetime, timedelta
from typing import List, Tuple

# import from package
from resistics.ioHandlers.ioHandler import IOHandler
from resistics.dataObjects.statisticData import StatisticData
from resistics.utilities.utilsIO import checkAndMakeDir
from resistics.utilities.utilsPrint import arrayToString


class StatisticIO(IOHandler):
    """Class for reading and writing statistic data

    Statistics are calculated out for measurement directories and are located in:
    project -> statData -> site -> meas -> specdir -> statdir

	Attributes
	----------
	dataPath : str
		Path to mask file directory

	Methods
	-------
	__init__(datapath)
		Initialise the statisticIO handler.
	read(statName, inc)
		Read in statistic statName at decimation level inc
	write(statData: StatisticData, inc: int)
		Write out statistic data at decimation level inc
	getFileNames(datapath, filename, inc)
		Get the filenames of the statistic data, the statistic info file and the statistic comments file in location datapath, with base name filename and decimation level inc
    printList()
        Class status returned as list of strings       
	"""

    def __init__(self, datapath="") -> None:
        """Initialise
        
        Parameters
        ----------
        datapath : str, optional
            The path to the statistic data
        """

        self.datapath: str = datapath

    def setDatapath(self, datapath: str) -> None:
        """Set the datapath

        Parameters
        ----------
        datapath : str
            The path to the statistic data
        """

        self.datapath = datapath

    def read(self, statName: str, inc: int) -> StatisticData:
        """Read a statistic file

        Parameters
        ----------
        statName : str
            The statistic name
        inc : int
            The increment, usually the decimation level
        
        Returns
        -------
        StatisticData
            A statistic data object
        """

        statFile, infoFile, commentFile = self.getFileNames(
            self.datapath, statName, inc
        )
        statFile = statFile + ".npy"
        # want a channel ordering
        if not os.path.exists(infoFile) or not os.path.exists(statFile):
            self.printWarning(
                "Unable to find info file {} or stat file {}".format(
                    infoFile, statFile
                ),
            )
            return False
        f = open(infoFile, "r")
        lines = f.readlines()
        f.close()
        refTime = datetime.strptime(lines[0].strip(), "%Y-%m-%d %H:%M:%S.%f")
        sampleFreq = float(lines[1].strip())
        winSize = int(lines[2].strip())
        winOlap = int(lines[3].strip())
        numWindows = int(lines[4].strip())
        winStats = lines[5].strip().split()
        evalFreq = np.fromstring(lines[6].strip(), dtype=float, sep=",")
        dtype = lines[7].strip()
        # now deal with the global indices
        indexInformation = lines[
            9:
        ]  # this is the information about local to global map
        globalIndicesAll = np.empty(shape=(numWindows), dtype=int)
        localIndices = []
        globalIndices = []
        for inInfo in indexInformation:
            inInfo = inInfo.strip()
            if inInfo == "":
                continue
            # now get the local and global indices
            split = inInfo.split("-")
            localIndices.append(int(split[0].strip()))
            globalIndices.append(int(split[1].strip()))
        # now fill the global indices array
        for idx, localI in enumerate(localIndices):
            if idx == len(localIndices) - 1:
                numWindowsFromLocal = numWindows - localI
                globalIndicesAll[localI:] = np.arange(
                    globalIndices[idx], globalIndices[idx] + numWindowsFromLocal
                )
                break
            numWindowsFromLocal = localIndices[idx + 1] - localI
            globalIndicesAll[localI : localI + numWindowsFromLocal] = np.arange(
                globalIndices[idx], globalIndices[idx] + numWindowsFromLocal
            )

        # load the statistics
        stats = np.load(statFile)

        # finally, read the comments file
        comments = []
        if os.path.exists(commentFile):
            f = open(commentFile, "r")
            comments = f.readlines()
            for idx, c in enumerate(comments):
                comments[idx] = c.rstrip()
            f.close()

        # now return a statisticData object
        return StatisticData(
            statName,
            refTime,
            sampleFreq,
            winSize,
            winOlap,
            numWindows=numWindows,
            winStats=winStats,
            evalFreq=evalFreq,
            stats=stats,
            globalIndices=globalIndicesAll,
            comment=comments,
            dtype=dtype,
        )

    def write(self, statData: StatisticData, inc: int) -> None:
        """Write a statistic file

        Parameters
        ----------
        statData : StatisticData
            The statistic data to write out
        inc : int
            The increment, usually the decimation level
        """

        # write the info file - numWindows and channel ordering
        checkAndMakeDir(os.path.join(self.datapath, "{}".format(statData.statName)))
        statFile, infoFile, commentFile = self.getFileNames(
            self.datapath, statData.statName, inc
        )

        # info file
        # want a channel ordering
        f = open(infoFile, "w")
        f.write("{}\n".format(statData.refTime.strftime("%Y-%m-%d %H:%M:%S.%f")))
        f.write("{}\n".format(statData.sampleFreq))
        f.write("{}\n".format(statData.winSize))
        f.write("{}\n".format(statData.winOlap))
        f.write("{:d}\n".format(statData.numWindows))
        f.write("{}\n".format(" ".join(statData.winStats)))
        # now write out the evaluation frequencies
        f.write("{}\n".format(arrayToString(statData.evalFreq)))
        # finally, write out the datatype
        f.write("{}\n".format(statData.dtype))
        # write out the window information
        # only write out when not consecutive
        f.write("Window map: localIndex - globalIndex\n")
        prevI = -10
        for idx, gIndex in enumerate(statData.globalIndices):
            if gIndex != prevI + 1:
                f.write("{:d} - {:d}\n".format(idx, gIndex))
            prevI = gIndex
        f.close()

        # check to save comments
        # only want this on the first increment - no need to rewrite after the others
        if inc < 1:
            comments = statData.comments + [
                "Statistic data for statistic {} written to {} on {}".format(
                    statData.statName, self.datapath, datetime.now()
                )
            ]
            f = open(commentFile, "w")
            for c in comments:
                f.write("{}\n".format(c))
            f.close()

        # save binary stat file
        np.save(statFile, statData.stats)

    def getFileNames(
        self, datapath: str, filename: str, inc: int
    ) -> Tuple[str, str, str]:

        """Get the statistic file name

        Parameters
        ----------
        datapath : str
            The datapath to the statistics data
        filename : str
            The filename of the statistics data
        inc : int
            The increment level, usually the decimation level
        
        Returns
        -------
        statFile : str
            Name of the statistics data file
        infoFile : str
            Name of the statistics info file
        commentFile : str
            Name of the statistics comments file
        """

        # inc is meant to be the decimation parameter
        # exclude extension on statfile because .npy added by numpy
        statFile = ""
        infoFile = ""
        if inc < 0:
            statFile = os.path.join(
                datapath, "{}".format(filename), "{}".format(filename)
            )
            infoFile = os.path.join(
                datapath, "{}".format(filename), "{}.info".format(filename)
            )
        else:
            statFile = os.path.join(
                datapath, "{}".format(filename), "{}{:02d}".format(filename, inc)
            )
            infoFile = os.path.join(
                datapath, "{}".format(filename), "{}{:02d}.info".format(filename, inc)
            )
        # name of the comment file
        commentFile = os.path.join(datapath, "{}".format(filename), "comments.txt")
        return statFile, infoFile, commentFile

    def printList(self) -> List[str]:  #
        """Get class information as a list of strings

        Returns
        -------
        List[str]
            Class information as a list of strings
        """

        textLst = []
        if self.datapath == "":
            textLst.append(
                "No datapath given. Set the datapath using the setDatapath function"
            )
        else:
            textLst.append("Datapath = {}".format(self.datapath))
        return textLst




import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.colors as colors
from matplotlib.dates import (
    DateFormatter,
    DayLocator,
    AutoDateLocator,
    AutoDateFormatter,
)
from datetime import datetime, timedelta
from typing import Dict, Tuple

# import from package
from resistics.utilities.utilsChecks import parseKeywords


def savePlot(path: str, filename: str, fig: plt.figure) -> str:
    """Save plot whilst checking for overwriting
    
    Parameters
    ----------
    path : str
        The path to save the file
    filename : str
        Filename to save to
    fig : plt.figure   
        Matplotlig figure to save
    
    Returns
    -------
    str
        The filename written out to
    """

    import os

    filenameForWriting = filename
    fileExists = True
    counter = 1
    while fileExists:
        filepath = os.path.join(path, "{}.png".format(filenameForWriting))
        if os.path.exists(filepath):
            filenameForWriting = "{}.{}".format(filename, counter)
            counter += 1
        else:
            fileExists = False
            break
    filepath = os.path.join(path, "{}.png".format(filenameForWriting))
    fig.savefig(filepath)
    return filenameForWriting


def getPlotRowsAndCols(maxCols: int, numPlots: int = 0) -> Tuple[int, int]:
    """Get the numbers of rows and columns for plots

    Parameters
    ----------
    maxCols : int
        The maximum number of columns
    numPlots : int
        The of subplots
    """

    if numPlots <= maxCols:
        ncols = numPlots
        nrows = 1
    else:
        ncols = maxCols
        nrows = int(np.ceil(1.0 * numPlots / maxCols))
    return nrows, ncols


def plotOptionsStandard(**kwargs) -> Dict:
    """Get a set of standard plot options

    Parameters
    ----------
    figsize : Tuple, optional
        Set the figure size
    plotfonts : Dict, optional
        Font sizes to use for plotting fonts
    block : bool, optional
        Boolean flag for blocking execution when plot is shown

    Returns
    -------
    Dict
        Dictionary of standard plot options
    """

    default: Dict = {}
    default["figsize"] = (20, 12)
    default["plotfonts"] = getViewFonts()
    default["block"] = True
    default = parseKeywords(default, kwargs)
    return default


def plotOptionsTime(**kwargs) -> Dict:
    """Get default plot options for plotting time data 

    Parameters
    ----------
    figsize : Tuple, optional
        Set the figure size
    plotfonts : Dict, optional
        Font sizes to use for plotting fonts
    block : bool, optional
        Boolean flag for blocking execution when plot is shown
    Eylim : List[float], optional
        y limits for electric data
    Hylim : List[float], optional
        y limits for magnetic data

    Returns
    -------
    out : Dict
        Dictionary of default plot options for plotting time data
    """

    default = plotOptionsStandard()
    default["Eylim"] = []
    default["Hylim"] = []
    default = parseKeywords(default, kwargs)
    return default


def plotOptionsSpec(**kwargs) -> Dict:
    """Get default plot options for plotting spectra data 

    Parameters
    ----------
    figsize : Tuple, optional
        Set the figure size
    plotfonts : Dict, optional
        Font sizes to use for plotting fonts
    block : bool, optional
        Boolean flag for blocking execution when plot is shown
    amplim : List[float], optional
        Amplitude limits for plotting spectra

    Returns
    -------
    out : Dict
        Dictionary of default plot options for plotting spectra data
    """

    default = plotOptionsStandard()
    default["amplim"] = []
    default = parseKeywords(default, kwargs)
    return default


def plotOptionsTransferFunction(**kwargs) -> Dict:
    """Get default plot options for plotting transfer function data 

    Parameters
    ----------
    figsize : Tuple, optional
        Set the figure size
    plotfonts : Dict, optional
        Font sizes to use for plotting fonts
    block : bool, optional
        Boolean flag for blocking execution when plot is shown
    res_ylim : List[float], optional
        y limits for resistivity data
    phase_ylim : List[float], optional
        y limits for phase data
    xlim : List[float], optional
        x limits for transfer function data

    Returns
    -------
    out : Dict
        Dictionary of default plot options for plotting spectra data
    """

    default = plotOptionsStandard()
    default["figsize"] = None
    default["res_ylim"] = [0.01, 10000]
    default["phase_ylim"] = [-20, 90]
    default["xlim"] = [0.0001, 10000]
    default = parseKeywords(default, kwargs)
    return default


def getTransferFunctionFigSize(oneplot: bool, npolarisations: int) -> Tuple[int, int]:
    """Get the plot size for a transfer function plot

    Notes
    -----
    This is awkward because of maintaining aspect ratio for the apparent resistivity. A better solution might be required at some point.

    Parameters
    ----------
    oneplot : bool
        All polarisations in one plot
    npolarisations : int
        Number of polarisations to plot
    
    Returns
    -------
    figsize : Tuple
        The figure size
    """

    if oneplot:
        return (6, 10)
    else:
        return (5 * npolarisations, 9)


def plotOptionsTipper(**kwargs) -> Dict:
    """Get default plot options for plotting transfer function data 

    Parameters
    ----------
    figsize : Tuple, optional
        Set the figure size
    plotfonts : Dict, optional
        Font sizes to use for plotting fonts
    block : bool, optional
        Boolean flag for blocking execution when plot is shown
    length_ylim : List[float], optional
        y limits for length data
    angle_ylim : List[float], optional
        y limits for angle data
    xlim : List[float], optional
        x limits for transfer function data

    Returns
    -------
    out : Dict
        Dictionary of default plot options for plotting spectra data
    """

    default = plotOptionsStandard()
    default["figsize"] = (16, 5)
    default["length_ylim"] = [0.001, 1000]
    default["angle_ylim"] = [-30, 30]
    default["xlim"] = [0.0001, 10000]
    default = parseKeywords(default, kwargs)
    return default


def getViewFonts() -> Dict:
    """Get default plot font options for viewing plots   

    Returns
    -------
    out : Dict
        Dictionary of default plot font options
    """

    plotFonts: Dict = {}
    plotFonts["suptitle"] = 14
    plotFonts["title"] = 12
    plotFonts["axisLabel"] = 12
    plotFonts["axisTicks"] = 12
    plotFonts["legend"] = 12
    return plotFonts


def getPlotFonts() -> Dict:
    """Get default plot font options for saving plots   

    Returns
    -------
    out : Dict
        Dictionary of default plot font options
    """

    plotFonts: Dict = {}
    plotFonts["suptitle"] = 18
    plotFonts["title"] = 16
    plotFonts["axisLabel"] = 16
    plotFonts["axisTicks"] = 14
    plotFonts["legend"] = 14
    return plotFonts


def getPresentationFonts() -> Dict:
    """Get default plot font options for presentation plots   

    Returns
    -------
    out : Dict
        Dictionary of default plot font options
    """

    plotFonts: Dict = {}
    plotFonts["suptitle"] = 22
    plotFonts["title"] = 20
    plotFonts["axisLabel"] = 20
    plotFonts["axisTicks"] = 20
    plotFonts["legend"] = 20
    return plotFonts


def getPaperFonts() -> Dict:
    """Get default plot font options for paper plots   

    Returns
    -------
    out : Dict
        Dictionary of default plot font options
    """

    plotFonts: Dict = {}
    plotFonts["suptitle"] = 18
    plotFonts["title"] = 17
    plotFonts["axisLabel"] = 16
    plotFonts["axisTicks"] = 15
    plotFonts["legend"] = 15
    return plotFonts


def transferFunctionColours() -> Dict[str, str]:
    """Get colours from the different components of standard magnetotelluric transfer functions
    
    Returns
    -------
    Dict[str, str]
        Dictionary mapping magnetotelluric transfer funciton components to colours
    """

    colours = {
        "ExHx": "orange",
        "EyHy": "green",
        "ExHy": "red",
        "EyHx": "blue",
        "HzHx": "gray",
        "HzHy": "magenta",
    }
    return colours


def colorbar2dTime():
    """Get default colormap for 2d plots with time   

    Returns
    -------
    out : plt.cm
        Colormap
    """

    return plt.cm.viridis


def colorbar2dSpectra():
    """Return colorbar for plotting spectra sections

    Returns
    -------
    out : plt.cm
        Colormap
    """

    return plt.cm.magma


def colorbar2dOther():
    """Alternative colormap for 2d data 

    Returns
    -------
    out : plt.cm
        Colormap
    """

    return plt.cm.plasma

def colorbarMultiline():
    """Alternative colormap for plotting multiple line plots 

    Returns
    -------
    out : plt.cm
        Colormap
    """

    return plt.cm.cividis    


def addLegends(fig: plt.figure, pos: int = 1) -> None:
    """Add legends to all subplots in figure
    
    Parameters
    ----------
    fig : magplotlib.pyplot.figure
        A matplotlib figure
    pos : int
        Location of legend  
    """

    axList = fig.axes
    for ax in axList:
        ax.legend(loc=pos)



import inspect
from datetime import datetime
from typing import Dict, List, Union, Any

# import from package
from resistics.dataObjects.configData import ConfigData
from resistics.ioHandlers.dataReader import DataReader
from resistics.calculators.calibrator import Calibrator
from resistics.calculators.decimationParameters import DecimationParams
from resistics.calculators.windowParameters import WindowParams
from resistics.calculators.windowSelector import WindowSelector
from resistics.calculators.processorSingleSite import ProcessorSingleSite
from resistics.calculators.processorRemoteReference import ProcessorRemoteReference
from resistics.dataObjects.timeData import TimeData
from resistics.dataObjects.projectData import ProjectData
from resistics.utilities.utilsPrint import (
    generalPrint,
    warningPrint,
    errorPrint,
    blockPrint,
)


def projectText(infoStr: str) -> None:
    """General print to terminal

    Parameters
    ----------
    infoStr : str
        The string to print to the console
    """

    generalPrint("{} Info".format(inspect.stack()[1][3]), infoStr)


def projectBlock(textLst: List[str]) -> None:
    """Class information as a list of strings

    Parameters
    ----------
    textLst : list[str]
        List of strings with information
    """

    blockPrint(inspect.stack()[1][3], textLst)


def projectWarning(warnStr: str) -> None:
    """Warning print to terminal

    Parameters
    ----------
    warnStr : str
        The string to print to the console
    """

    warningPrint("{} Warning".format(inspect.stack()[1][3]), warnStr)


def projectError(errorStr: str, quitRun: bool = False) -> None:
    """Error print to terminal and possibly quit

    Parameters
    ----------
    errorStr : str
        The string to print to the console
    quitRun : bool, optional (False)
        If True, the code will exit
    """

    errorPrint("{} Warning".format(inspect.stack()[1][3]), errorStr, quitRun=quitRun)


def getCalibrator(calPath: str, config: Union[ConfigData, None] = None) -> Calibrator:
    """Create a Calibnator object from calibration path and configuration information

    Parameters
    ----------
    calPath : str
        Path to calibration directory
    config : ConfigData
        Configuration data
    
    Returns
    -------
    Calibrator
        A calibrator object
    """

    cal = Calibrator(calPath)
    if config is None:
        return cal

    cal.useTheoretical = config.configParams["Calibration"]["usetheoretical"]
    return cal


def getDecimationParameters(
    sampleFreq: float, config: Union[ConfigData, None] = None
) -> DecimationParams:
    """Create a DecimationParams object from sampling frequency and configuration information

    Parameters
    ----------
    sampleFreq : float
        Sampling frequency of the data
    config : ConfigData
        Configuration data

    Returns
    -------
    DecimationParams
        A decimation parameters object        
    """

    decParams = DecimationParams(sampleFreq)
    if config is None:
        return decParams

    if config.flags["customfrequencies"]:
        decParams.setFrequencyParams(
            config.configParams["Frequencies"]["frequencies"],
            config.configParams["Decimation"]["numlevels"],
            config.configParams["Frequencies"]["perlevel"],
        )
    else:
        decParams.setDecimationParams(
            config.configParams["Decimation"]["numlevels"],
            config.configParams["Frequencies"]["perlevel"],
        )

    return decParams


def getWindowParameters(
    decParams: DecimationParams, config: Union[ConfigData, None] = None
) -> WindowParams:
    """Create a WindowParams object from decimationParams and configuration data

    Parameters
    ----------
    decParams : DecimationParams
        DecimationParams object to hold the decimation parameters
    config : ConfigData
        Configuration data
    
    Returns
    -------
    WindowParams
        A window parameters object        
    """

    winParams = WindowParams(decParams)
    if config is None:
        return winParams

    if config.flags["customwindows"]:
        winParams.setWindowParameters(
            config.configParams["Window"]["windowsizes"],
            config.configParams["Window"]["overlapsizes"],
        )
    else:
        winParams.setMinParams(
            config.configParams["Window"]["minwindowsize"],
            config.configParams["Window"]["minoverlapsize"],
        )

    return winParams


def getWindowSelector(
    projData: ProjectData,
    decParams: DecimationParams,
    winParams: WindowParams,
) -> WindowSelector:
    """Create a WindowSelector object from projectData, decimationParams, windowParams

    Parameters
    ----------
    projData : ProjectData
        A project data object
    decParams : DecimationParams
        DecimationParams object to hold the decimation parameters
    winParams : Windowparams
        WindowParams object to hold the windowing parameters
    
    Returns
    -------
    WindowSelector
        A window selector object        
    """

    selector = WindowSelector(
        projData,
        decParams.sampleFreq,
        decParams,
        winParams,
        specdir=projData.config.configParams["Spectra"]["specdir"],
    )
    return selector


def getSingleSiteProcessor(
    winSelector: WindowSelector, outPath: str, config: Union[ConfigData, None] = None
) -> ProcessorSingleSite:
    """Create a ProcessorSingleSite object from a windowSelector object, outPath and config data

    Parameters
    ----------
    winSelector : WindowSelector
        Window selector with sites, masks etc already specified
    tfPath : str
        Path to output transfer function data
    config : ConfigData, optional
        Configuration data
    
    Returns
    -------
    ProcessorSingleSite
        A single site processor        
    """

    processor = ProcessorSingleSite(winSelector, outPath)
    if config is None:
        return processor

    return processor


def getRemoteReferenceProcessor(
    winSelector: WindowSelector, outPath: str, config: Union[ConfigData, None] = None
):
    """Create a ProcessorRemoteReference object from a windowSelector object, outPath and config data

    Parameters
    ----------
    winSelector : WindowSelector
        Window selector with sites, masks etc already specified
    outPath : str
        Path to output transfer function data
    config : ConfigData, optional
        Configuration data
    
    Returns
    -------
    ProcessorRemoteReference
        A remote reference processor        
    """

    processor = ProcessorRemoteReference(winSelector, outPath)
    if config is None:
        return processor

    return processor


def checkDateOptions(options: Dict, timeStart: datetime, timeStop: datetime) -> bool:
    """Check to see if data contributes to user specified date range

    Parameters
    ----------
    options : Dict
        Options dictionary with start and stop options specified by user (if specified at all)
    timeStart : datetime
        Start time of data
    timeStop : datetime
        Stop time of data

    Returns
    -------
    bool
        True if data contributes to the date range
    """

    # now check the user provided dates
    if options["start"] and options["start"] > timeStop:
        # this data has nothing to contribute in the optional date range
        return False
    if options["stop"] and options["stop"] < timeStart:
        # this data has nothing to contribute in the optional date range
        return False
    return True


def applyPolarisationReversalOptions(options: Dict, timeData: TimeData) -> TimeData:
    """Polarity reverse time data

    Parameters
    ----------
    options : Dict
        User specified options for polarity reversal
    timeData : TimeData
        Time data to polarity reverse

    Returns
    -------
    TimeData
        Polarity reversed time data
    """

    if isinstance(options["polreverse"], bool):
        # no polarity reversal to be performed
        return timeData

    if not isinstance(options["polreverse"], dict):
        # not specified in the right way
        return timeData

    from resistics.utilities.utilsMath import polarityReversal

    timeData = polarityReversal(timeData, options["polreverse"])
    return timeData


def applyCalibrationOptions(
    options: Dict, cal: Calibrator, timeData: TimeData, reader: DataReader
) -> TimeData:
    """Calibrate time data with user options

    To calibrate, specify
    options["calibrate"] = True

    Parameters
    ----------
    options : Dict
        User specified options for calibrating
    cal : Calibrator
        A calibrator instance
    timeData : TimeData
        Time data to filter
    reader: DataReader
        A data reader object for the data

    Returns
    -------
    TimeData
        Calibrated time data
    """

    if options["calibrate"]:
        sensors = reader.getSensors(timeData.chans)
        serials = reader.getSerials(timeData.chans)
        choppers = reader.getChoppers(timeData.chans)
        timeData = cal.calibrate(timeData, sensors, serials, choppers)
    return timeData


def applyFilterOptions(options: Dict, timeData: TimeData) -> TimeData:
    """Filter time data

    To low pass filter, specify
    options["filter"]["lpfilt"] = high cutoff frequency 
    To high pass filter, specify 
    options["filter"]["hpfilt"] = low cutoff frequency
    To bandass filter, specify
    options["filter"]["bpfilt"] = [low cutoff frequency, high cutoff frequency]

    Parameters
    ----------
    options : Dict
        User specified options for filtering
    timeData : TimeData
        Time data to filter

    Returns
    -------
    TimeData
        Filtered time data
    """

    if "lpfilt" in options["filter"]:
        from resistics.utilities.utilsFilter import lowPass

        timeData = lowPass(timeData, options["filter"]["lpfilt"])
    if "hpfilt" in options["filter"]:
        from resistics.utilities.utilsFilter import highPass

        timeData = highPass(timeData, options["filter"]["hpfilt"])
    if "bpfilt" in options["filter"]:
        from resistics.utilities.utilsFilter import bandPass

        timeData = bandPass(
            timeData, options["filter"]["bpfilt"][0], options["filter"]["bpfilt"][1]
        )
    return timeData


def applyNotchOptions(options: Dict, timeData: TimeData) -> TimeData:
    """Notch filter time data

    To notch filter, options["notch"] needs to be a list of frequencies to filter out. For example, to notch filter at 16.6Hz, this would be:
    options["notch"] = [16.6]
    For two frequencies, for example 16.6Hz and 50Hz, then:
    options["notch"] = [16.6, 50]

    Parameters
    ----------
    options : Dict
        User specified options for notching
    timeData : TimeData
        Time data to filter

    Returns
    -------
    TimeData
        Notch filtered time data
    """

    from resistics.utilities.utilsFilter import notchFilter

    if len(options["notch"]) > 0:
        for n in options["notch"]:
            timeData = notchFilter(timeData, n)
    return timeData


def applyNormaliseOptions(options: Dict, timeData: TimeData) -> TimeData:
    """Normalise time data

    To normalise, options["normalise"] needs to be set to True

    Parameters
    ----------
    options : Dict
        User specified options for normalising
    timeData : TimeData
        Time data to filter

    Returns
    -------
    TimeData
        Normalised time data
    """

    from resistics.utilities.utilsFilter import normalise

    if options["normalise"]:
        projectText("Normalising data")
        timeData = normalise(timeData)
    return timeData


def applyInterpolationOptions(options: Dict, timeData: TimeData) -> TimeData:
    """Interpolate time data

    Interpolate time data to make sure all the data starts on a full second. This is best to do before resampling.

    To interpolate, options["interp"] needs to be set to True

    Parameters
    ----------
    options : Dict
        Interpolation options in a dictionary
    timeData : TimeData
        Time data object
    
    Returns
    -------
    TimeData
        Interpolated time data

    Notes
    -----
    This will fail with longer sample periods (i.e. greater than a second)    
    """

    from resistics.utilities.utilsInterp import interpolateToSecond

    if options["interp"]:
        if timeData.startTime.microsecond != 0:
            timeData = interpolateToSecond(timeData)
    return timeData


def applyResampleOptions(options: Dict, timeData: TimeData) -> TimeData:
    """Resample time data

    The resampling options in the options dictionary should be specified as:
    options["resample"][frequency to resample from] = frequency to resample to

    Parameters
    ----------
    options : Dict
        Interpolation options in a dictionary
    timeData : TimeData
        Time data object
    
    Returns
    -------
    TimeData
        Resampled time data
    """

    from resistics.utilities.utilsFilter import resample

    if timeData.sampleFreq in options["resamp"]:  # then need to resample this data
        timeData = resample(timeData, options["resamp"][timeData.sampleFreq])
    return timeData



import os
import glob
import xml.etree.ElementTree as ET
from datetime import datetime, timedelta
import numpy as np

# import from package
from resistics.ioHandlers.dataReader import DataReader


class DataReaderATS(DataReader):
    """Data reader for ATS formatted data
    
    For ATS files, header information is XML formatted. The end time in ATS header files is actually one sample past the time of the last sample. The dataReader handles this and gives an end time corresponding to the actual time of the last sample.

    Methods
    -------
    setParameters()
        Set data format parameters
    dataHeaders()
        Headers to read in
    readHeaders()
        Specific function for reading the headers for internal format
    lineToKeyAndValue(line)
        Separate a line into key and value with = as a delimiter

    Notes
    -----
    The raw data units for ATS data are in counts. To get data in field units, ATS data is first multipled by the least significat bit (lsb) defined in the header files,

    .. code-block:: text  
    
        data = data * leastSignificantBit,
    
    giving data in mV. The lsb includes the gain removal, so no separate gain removal needs to be performed.
    
    For electrical channels, there is additional step of dividing by the electrode spacing, which is provided in metres. The extra factor of a 1000 is to convert this to km to give mV/km for electric channels
    
    .. code-block:: text  
        
        data = (1000 * data)/electrodeSpacing
    
    Finally, to get magnetic channels in nT, the magnetic channels need to be calibrated.
    """

    def setParameters(self) -> None:
        """Set data reader parameters for ATS files"""

        self.headerF = glob.glob(os.path.join(self.dataPath, "*.xml"))
        self.dataF = glob.glob(os.path.join(self.dataPath, "*.ats"))
        self.dataByteOffset = 1024
        self.dataByteSize = 4

    def dataHeaders(self):
        """Return the data headers in the internal file format
        
        Returns
        -------
        recordingHeaders : List[str]
            Headers with information about the recording
        globalHeaders : List[str]
            Common headers with information about the recording
        channelHeadersInput : List[str]
            Channel setup headers
        channelHeadersOutput : List[str]
            Channel recording headers      
        """

        recordingHeaders = ["start_time", "start_date", "stop_time", "stop_date"]
        globalHeaders = ["meas_channels", "sample_freq"]
        channelHeadersInput = ["gain_stage1", "gain_stage2", "hchopper", "echopper"]
        channelHeadersOutput = [
            "start_time",
            "start_date",
            "sample_freq",
            "num_samples",
            "ats_data_file",
            "sensor_type",
            "channel_type",
            "ts_lsb",
            "pos_x1",
            "pos_x2",
            "pos_y1",
            "pos_y2",
            "pos_z1",
            "pos_z2",
            "sensor_sernum",
        ]
        return (
            recordingHeaders,
            globalHeaders,
            channelHeadersInput,
            channelHeadersOutput,
        )

    def readHeader(self):
        """Read time data header file for ATS format
        
        Headers for ATS files are XML formatted.
        """

        if len(self.headerF) > 1:
            self.printWarning(
                "More xml files than expected. Using: {}".format(self.headerF[0])
            )
        tree = ET.parse(self.headerF[0])
        root = tree.getroot()
        # get header names
        rHeaders, gHeaders, cHeadersInput, cHeadersOutput = self.dataHeaders()

        # get recording headers
        self.headers = {}
        recording = root.find("./recording")
        for rH in rHeaders:
            self.headers[rH] = recording.find(rH).text
        # get global config headers
        globalConfig = recording.find("./input/ADU07Hardware/global_config")
        for gH in gHeaders:
            self.headers[gH] = globalConfig.find(gH).text

        # get the channel headers in the input section
        self.chanHeaders = []
        for chan in root.findall(
            "./recording/input/ADU07Hardware/channel_config/channel"
        ):
            chanH = {}
            for cH in cHeadersInput:
                chanH[cH] = chan.find(cH).text
            self.chanHeaders.append(chanH)
        # get the channel headers in the output section
        outputSec = root.findall(
            "./recording/output/ProcessingTree1/output/ATSWriter/configuration/channel"
        )
        # check for old style xml file
        if len(outputSec) == 0:
            outputSec = root.findall(
                "./recording/output/ATSWriter/configuration/channel"
            )
        for chan, chanH in zip(outputSec, self.chanHeaders):
            for cH in cHeadersOutput:
                chanH[cH] = chan.find(cH).text

        # a couple of things to do: add microseconds to the times
        # remember, the actual end time is one sample back
        # if you do a calculation with the number of samples and the start time
        self.headers["start_time"] = self.headers["start_time"] + ".000000"
        self.headers["stop_time"] = self.headers["stop_time"] + ".000000"
        for chanH in self.chanHeaders:
            chanH["start_time"] = chanH["start_time"] + ".000000"

        # set the lsb applied header in chans
        # for ats files, this is not applied in the raw data files
        for idx, ch in enumerate(self.chanHeaders):
            self.chanHeaders[idx]["scaling_applied"] = False
            self.chanHeaders[idx]["ts_lsb"] = "-{}".format(
                self.chanHeaders[idx]["ts_lsb"]
            )



import os
import glob
import re, struct
import collections
import copy
from datetime import datetime, timedelta
import numpy as np
from typing import List, Dict, Tuple

# import from package
from resistics.ioHandlers.dataReader import DataReader
from resistics.ioHandlers.dataWriterInternal import DataWriterInternal
from resistics.dataObjects.timeData import TimeData
from resistics.utilities.utilsChecks import consistentChans, isMagnetic, isElectric
from resistics.utilities.utilsClean import (
    removeZeros,
    removeZerosSingle,
    removeNansSingle,
)
from resistics.utilities.utilsMath import intdiv
from resistics.utilities.utilsPrint import blockPrint


class DataReaderPhoenix(DataReader):
    """Data reader for Phoenix data

    The Phoenix data and recording format is different and does not nicely fit with the way resistics tries to model data.
    
    There are three frequencies recorded concurrently (e.g. 2400Hz, 150Hz, 15Hz). The lowest sampling frequency is continuous whilst the others record data files at regular intervals. There is no issue with the continous sampling frequency. 
    
    However, as resistics separates out data into continuous recordings, the consistent gaps for the higher frequencies will lead to lots of small data folders if converted to internal data format.

    This class returns the lowest frequency recording (the continuous one) when time series data is requested. However, higher frequencies can be converted to the internal data format using the methods available here.

    Warnings
    --------
    The appropriate scaling for Phoenix data to return field units has not yet been verified.

    It is not actually recommended to reformat the high frequency recordings as this will lead to potentially thousands of data folders. There is currently no straight-forward way to support the high-frequency Phoenix recordings.

    Attributes
    ----------
    recChannels : Dict
        Channels in each data file
    dtype : np.float32
        The data type
    numHeaderFiles : int
        The number of header files
    numDataFiles : int
        The number of data files

    Methods
    -------
    setParameters()
        Set data reader parameters for Phoenix files
    getSamplesRatesTS()
        Get the sampling frequencies of the time series data
    getNumberSamplesTS()
        Get the number of samples for each time series file
    getUnscaledSamples(**kwargs)
        Get raw data from data file
    getRecordsForSamples(startSample, endSample)
        Get the records to read for a sample range
    readTag(dataFile)
        Read the tag from a data file
    readRecord(dataFile, numChans, numScans)
        Read numScans from a record
    twosComplement(dataBytes)
        Read the two's complement data from the file
    getPhysicalSamples(**kwargs)
        Get data scaled to physical values
    chanDefaults()
        Get defaults for channel headers
    readHeader()
        Read header file
    readTable()
        Read table file
    removeControl(inBytes)
        Remove control characters from a byte string
    headersFromTable(tableData)
        Parse the information in the table file to get headers
    getDates(tableData)
        Get recording dates (start and end time)
    checkSamples()
        Check the number of samples for all the timeseries (ts) files
    reformatHigh(path, **kwargs)
        Write out high frequency time series in internal format
    reformatContinuous(path)
        Write out the continuous time series in internal format
    reformat(path)
        Write out all recorded time series to internal format
    printDataFileList()  
        Information about the data files as a list of strings
    printDataFileInfo()
        Print a list of the data files
    printTableFileList()
        Information about the table file as a list of strings
    printTableFileInfo()
        Print table file info

    Notes
    -----
    Phoenix data is stored in 3 bytes two's-complement format.
    """

    def setParameters(self) -> None:
        """Set data reader parameters for Phoenix files
        
        Phoenix time series data is not contiguous in the file and is separated into records. There are multiple time series data files, one for the continuous recording and two others for the other frequencies. Therefore, there are a few other class variables defined here than in the parent DataReader class.
        """

        # get a list of the header and data files in the folder
        self.headerF = glob.glob(os.path.join(self.dataPath, "*.TBL"))
        self.dataF = glob.glob(os.path.join(self.dataPath, "*.TS*"))
        # set the sample byte size
        self.sampleByteSize = 3  # two's complement
        self.tagByteSize = 32
        self.dtype = int
        # there will be multiple TS files in here
        # need to figure out
        self.numHeaderFiles = len(self.headerF)
        self.numDataFiles = len(self.dataF)

    def getSamplesRatesTS(self) -> Dict:
        """Get the sampling frequencies of the time series data

        Returns
        -------
        Dict
            Dictionary with the time series file number as keys and their sampling frequencies in Hz as values
        """

        info: Dict = {}
        for num, sr in zip(self.tsNums, self.tsSampleFreqs):
            info[num] = sr
        return info

    def getNumberSamplesTS(self) -> Dict:
        """Get the number of samples for each time series file

        Returns
        -------
        Dict
            Dictionary with the time series file number as keys and their number of samples as values
        """

        info = {}
        for num, ns in zip(self.tsNums, self.tsNumSamples):
            info[num] = ns
        return info

    def getUnscaledSamples(self, **kwargs) -> TimeData:
        """Get raw data from data file

        Only returns the continuous data. The continuous data is in 24 bit two's complement (3 bytes) format and is read in using struct as this is not supported by numpy.
        
        Parameters
        ----------
        chans : List[str], optional
            List of channels to return if not all are required
        startSample : int, optional
            First sample to return
        endSample : int, optional
            Last sample to return

        Returns
        -------
        TimeData
            Time data object
        """

        # initialise chans, startSample and endSample with the whole dataset
        options = self.parseGetDataKeywords(kwargs)

        # get the files to read and the samples to take from them, in the correct order
        recordsToRead, samplesToRead = self.getRecordsForSamples(
            options["startSample"], options["endSample"]
        )
        numSamples = options["endSample"] - options["startSample"] + 1
        # set up the dictionary to hold the data
        data = {}
        for chan in options["chans"]:
            data[chan] = np.zeros(shape=(numSamples), dtype=self.dtype)

        # open the file
        dFile = open(self.continuousF, "rb")

        # loop through chans and get data
        sampleCounter = 0
        for record, sToRead in zip(recordsToRead, samplesToRead):
            # number of samples to read in record
            dSamples = sToRead[1] - sToRead[0] + 1
            # find the byte read start and byte read end
            recordByteStart = self.recordBytes[self.continuous][record]
            recordSampleStart = self.recordSampleStarts[self.continuous][record]
            # find the offset on the readFrom bytes
            # now recall, each sample is recorded as a scan (all channels recorded at the same time)
            # so multiply by number of channels to get the number of bytes to read
            byteReadStart = (
                recordByteStart
                + (sToRead[0] - recordSampleStart)
                * self.sampleByteSize
                * self.getNumChannels()
            )
            bytesToRead = dSamples * self.sampleByteSize * self.getNumChannels()
            # read the data - numpy does not support 24 bit two's complement (3 bytes) - hence use struct
            dFile.seek(byteReadStart, 0)  # seek to start byte from start of file
            dataBytes = dFile.read(bytesToRead)
            dataRead = self.twosComplement(dataBytes)
            # now need to unpack this
            for chan in options["chans"]:
                # check to make sure channel exists
                self.checkChan(chan)
                # get the channel index - the chanIndex should give the right order in the data file
                # as it is the same order as in the header file
                chanIndex = self.chanMap[chan]
                # now populate the channel data appropriately
                data[chan][sampleCounter : sampleCounter + dSamples] = dataRead[
                    chanIndex : dSamples * self.getNumChannels() : self.getNumChannels()
                ]
            # increment sample counter
            sampleCounter = sampleCounter + dSamples  # get ready for the next data read
        # close file
        dFile.close()

        # return data
        startTime, stopTime = self.sample2time(
            options["startSample"], options["endSample"]
        )
        comment = "Unscaled data {} to {} read in from measurement {}, samples {} to {}".format(
            startTime,
            stopTime,
            self.dataPath,
            options["startSample"],
            options["endSample"],
        )
        return TimeData(
            sampleFreq=self.getSampleFreq(),
            startTime=startTime,
            stopTime=stopTime,
            data=data,
            comments=comment,
        )

    def getRecordsForSamples(
        self, startSample: int, endSample: int
    ) -> Tuple[List, List]:
        """Get the records to read for a sample range

        Parameters
        ----------
        startSample : int
            The starting sample of the range
        endSample : int
            The ending sample of the range
        
        Returns
        -------
        recordsToRead : List
            The records to read from the time series data files
        samplesToRead : List
            The samples to read from each record
        """

        recordsToRead = []
        samplesToRead = []
        for record, timeStart in enumerate(self.recordStarts[self.continuous]):
            recordStartSamp = self.recordSampleStarts[self.continuous][record]
            recordEndSamp = self.recordSampleStops[self.continuous][record]
            if recordStartSamp > endSample or recordEndSamp < startSample:
                continue  # nothing to read from this file
            # in this case, there is some overlap with the samples to read
            recordsToRead.append(record)
            readFrom = recordStartSamp  # i.e. the first sample in the datafile
            readTo = recordEndSamp  # this the last sample in the file
            if recordStartSamp < startSample:
                readFrom = startSample
            if recordEndSamp > endSample:
                readTo = endSample
            # this is an inclusive number readFrom to readTo including readTo
            samplesToRead.append([readFrom, readTo])
        return recordsToRead, samplesToRead

    def readTag(self, dataFile) -> Tuple[str]:
        """Read the tag from a data file

        Parameters
        ----------
        dataFile : file handle
            File handle of the data file
        
        Returns
        -------
        numScans : List
            Number of scans in the tag
        numChans : List
            Number of channels in the tag
        dateString : str
            The dataString of the tag
        """

        second = struct.unpack("b", dataFile.read(1))[0]
        minute = struct.unpack("b", dataFile.read(1))[0]
        hour = struct.unpack("b", dataFile.read(1))[0]
        day = struct.unpack("b", dataFile.read(1))[0]
        month = struct.unpack("b", dataFile.read(1))[0]
        year = struct.unpack("b", dataFile.read(1))[0]
        dayOfWeek = struct.unpack("b", dataFile.read(1))[0]
        century = struct.unpack("b", dataFile.read(1))[0]
        dateString = "{:02d}{:02d}-{:02d}-{:02d} {:02d}:{:02d}:{:02d}.000".format(
            century, year, month, day, hour, minute, second
        )
        # serial number
        serialNum = struct.unpack("h", dataFile.read(2))
        # num scans
        numScans = struct.unpack("h", dataFile.read(2))[0]
        # channels per scan
        numChans = struct.unpack("b", dataFile.read(1))[0]
        # tag length
        tagLength = struct.unpack("b", dataFile.read(1))
        # status code
        statusCode = struct.unpack("b", dataFile.read(1))
        # bit-wise saturation flags
        saturationFlag = struct.unpack("b", dataFile.read(1))
        # reserved
        reserved = struct.unpack("b", dataFile.read(1))
        # sample length
        sampleLength = struct.unpack("b", dataFile.read(1))
        # sample rate
        sampleRate = struct.unpack("h", dataFile.read(2))
        # units of sample rate: 0 = Hz, 1 = minute, 2 = hour, 3 = day
        sampleUnits = struct.unpack("b", dataFile.read(1))
        # clock status
        clockStatus = struct.unpack("b", dataFile.read(1))
        # clock error in micro seconds
        clockError = struct.unpack("i", dataFile.read(4))
        # reserved
        res1 = struct.unpack("b", dataFile.read(1))
        res2 = struct.unpack("b", dataFile.read(1))
        res3 = struct.unpack("b", dataFile.read(1))
        res4 = struct.unpack("b", dataFile.read(1))
        res5 = struct.unpack("b", dataFile.read(1))
        res6 = struct.unpack("b", dataFile.read(1))
        # returnt the important variables
        return numScans, numChans, dateString

    def readRecord(self, dataFile, numChans, numScans):
        """Read numScans from a record

        Parameters
        ----------
        dataFile : file handle
            File handle of the data file
        numScans : List
            Number of scans in the tag
        numChans : List
            Number of channels in the tag

        Returns
        -------
        data : np.ndarray(int)
            Record data
        """

        data = np.zeros(shape=(numChans, numScans), dtype="int")
        for scan in range(0, numScans):
            for chan in range(0, numChans):
                dataBytes = dataFile.read(3)
                data[chan, scan] = self.twosComplement(dataBytes)
        return data

    def twosComplement(self, dataBytes):
        """Read the two's complement data from the file

        This parses two's complement 24-bit integer, little endian, unsigned and signed. The method is to pad out 3 bytes out with a null byte and read as unsigned integer with little endian (<).        

        Parameters
        ----------
        dataByes : bytes
            The bytes to parse

        Returns
        -------
        data : np.ndarray(int)
            Record data
        """

        if len(dataBytes) % self.sampleByteSize != 0:
            self.printError(
                "The number of bytes divided by the sample byte size does not give an exact number",
                quitRun=True,
            )
        # calculate num samples, this should be exact
        numSamples = intdiv(len(dataBytes), self.sampleByteSize)
        dataRead = np.zeros(shape=(numSamples), dtype=self.dtype)
        for i in range(0, numSamples):
            sampleBytes = dataBytes[
                i * self.sampleByteSize : (i + 1) * self.sampleByteSize
            ]
            unsigned = struct.unpack("<I", sampleBytes + b"\x00")[0]
            signed = unsigned if not (unsigned & 0x800000) else unsigned - 0x1000000
            dataRead[i] = signed
        return dataRead

    def getPhysicalSamples(self, **kwargs) -> TimeData:
        """Get data scaled to physical values

        Parameters
        ----------
        chans : List[str]
            List of channels to return if not all are required
        startSample : int
            First sample to return
        endSample : int
            Last sample to return
        remaverage : bool
            Remove average from the data
        remzeros : bool
            Remove zeroes from the data
        remnans: bool
            Remove NanNs from the data

        Returns
        -------
        TimeData
            Time data object
        """

        options = self.parseGetDataKeywords(kwargs)
        # get data
        timeData = self.getUnscaledSamples(
            chans=options["chans"],
            startSample=options["startSample"],
            endSample=options["endSample"],
        )
        # need to remove the gain
        for chan in options["chans"]:
            # remove the gain
            timeData.data[chan] = 1.0 * timeData.data[chan] / self.getChanGain1(chan)
            timeData.addComment(
                "Scaling channel {} with scalar {} to give mV".format(
                    chan, 1.0 / self.getChanGain1(chan)
                )
            )

            # divide by distance in km
            if chan == "Ex":
                # multiply by 1000/self.getChanDx same as dividing by dist in km
                timeData.data[chan] = 1000 * timeData.data[chan] / self.getChanDx(chan)
                timeData.addComment(
                    "Dividing channel {} by electrode distance {} km to give mV/km".format(
                        chan, self.getChanDx(chan) / 1000.0
                    )
                )
            if chan == "Ey":
                # multiply by 1000/self.getChanDy same as dividing by dist in km
                timeData.data[chan] = 1000 * timeData.data[chan] / self.getChanDy(chan)
                timeData.addComment(
                    "Dividing channel {} by electrode distance {} km to give mV/km".format(
                        chan, self.getChanDy(chan) / 1000.0
                    )
                )

            # if remove zeros - False by default
            if options["remzeros"]:
                timeData.data[chan] = removeZerosSingle(timeData.data[chan])
            # if remove nans - False by default
            if options["remnans"]:
                timeData.data[chan] = removeNansSingle(timeData.data[chan])
            # remove the average from the data - True by default
            if options["remaverage"]:
                timeData.data[chan] = timeData.data[chan] - np.average(
                    timeData.data[chan]
                )

        # add comments
        timeData.addComment(
            "The required Phoneix scaling to field units is still unverified. This is experimental and use cautiously."
        )
        timeData.addComment(
            "Remove zeros: {}, remove nans: {}, remove average: {}".format(
                options["remzeros"], options["remnans"], options["remaverage"]
            )
        )
        return timeData

    def chanDefaults(self):
        """Get defaults for channel headers

        Returns
        -------
        Dict[str, Any]
            Dictionary of headers for channels and default values
        """

        chanH = {}
        chanH["gain_stage1"] = 1
        chanH["gain_stage2"] = 1
        chanH["hchopper"] = 0  # this depends on sample frequency
        chanH["echopper"] = 0
        # channel output information (sensor_type, channel_type, ts_lsb, pos_x1, pos_x2, pos_y1, pos_y2, pos_z1, pos_z2, sensor_sernum)
        chanH["ats_data_file"] = ""
        chanH["num_samples"] = 0
        chanH["sensor_type"] = ""
        chanH["channel_type"] = ""
        chanH["ts_lsb"] = 1
        chanH["scaling_applied"] = False
        chanH["pos_x1"] = 0
        chanH["pos_x2"] = 0
        chanH["pos_y1"] = 0
        chanH["pos_y2"] = 0
        chanH["pos_z1"] = 0
        chanH["pos_z2"] = 0
        chanH["sensor_sernum"] = 0
        return chanH

    def readHeader(self):
        """Read header file

        For phoenix data, the header file is the table file and it is binary formatted.
        """

        # first, find which ts files are available (2,3,4,5)
        # and the continuous recording frequency (the max)
        self.tsNums = []
        for tsfile in self.dataF:
            self.tsNums.append(int(tsfile[-1]))
        self.continuous = max(self.tsNums)
        self.continuousI = self.tsNums.index(self.continuous)
        self.continuousF = self.dataF[self.continuousI]
        # read the table data
        self.tableData = self.readTable()
        # and then populate the headers
        self.headers, self.chanHeaders = self.headersFromTable(self.tableData)
        # finally, check the number of samples in each file
        self.checkSamples()

    def readTable(self) -> Dict:
        """Read a header table

        Returns
        -------
        OrderedDict
            An ordered dictionary of header table data
        """

        if len(self.headerF) > 1:
            self.printWarning(
                "More table files than expected. Using: {}".format(self.headerF[0])
            )
        numBytes = os.path.getsize(self.headerF[0])
        tableFile = open(self.headerF[0], "rb")
        tableData = collections.OrderedDict()
        # loop through file and read
        bytesRead = 0
        headerWordSize = 4
        headerSize = 12
        dataSize = 13
        increment = headerSize + dataSize
        while bytesRead <= numBytes - increment:
            # formats for reading in
            # integers
            ints = [
                "SGIN",
                "EGNC",
                "HGNC",
                "EGN",
                "HGN",
                "ACDC",
                "ACDH",
                "V5SR",
                "MTSR",
                "LCHP",
                "L2NS",
                "L3NS",
                "L4NS",
                "DDAT",
                "TXPR",
                "TBVO",
                "TBVI",
                "INIT",
                "RQST",
                "MODE",
                "XDOS",
                "ATYP",
                "FNAM",
                "FLEN",
                "AQST",
                "HSMP",
                "CALS",
                "CCLS",
                "TEMP",
                "TMAX",
                "GFPG",
                "FFPG",
                "DSP",
                "CHEX",
                "CHEY",
                "CHHX",
                "CHHY",
                "CHHZ",
                "TCHN",
                "POTS",
                "NREF",
                "CCLT",
                "PZLT",
                "NSAT",
                "OCTR",
                "CLST",
                "TALS",
                "TCMB",
                "TERR",
                "LPFR",
                "LFRQ",
                "SNUM",
                "MXSC",
                "BADR",
                "NOBF",
                "SATR",
                "BAT1",
                "BAT2",
                "BAT3",
                "EXR",
                "EYR",
                "ELEV",
                "SRL2",
                "SRL3",
                "SRL4",
                "SRL5",
                "DISK",
                "STDE",
                "TOTL",
                "STDH",
            ]
            # UTC
            ints1_8 = [
                "TDSP",
                "LFIX",
                "TSYN",
                "STIM",
                "ETIM",
                "HTIM",
                "ETMH",
                "NUTC",
                "FTIM",
                "LTIM",
            ]
            # non-integer headers
            doubles = [
                "EXAC",
                "EXDC",
                "EYAC",
                "EYDC",
                "HXAC",
                "HXDC",
                "HYAC",
                "HYDC",
                "HZAC",
                "HZDC",
                "DXAC",
                "DXDC",
                "DYAC",
                "DYDC",
                "EXNR",
                "EXPR",
                "EYNR",
                "EYPR",
                "GNDR",
                "MAXR",
                "EAZM",
                "HAZM",
                "DECL",
                "TSTV",
                "FSCV",
                "CCMN",
                "CCMX",
                "HATT",
                "HAMP",
                "CPHC",
                "LFIX",
                "EXLN",
                "EYLN",
                "TSTR",
                "INPR",
                "CFMN",
                "CFMX",
                "HNOM",
            ]
            # get the header word
            header = struct.unpack(
                "{}s".format(headerWordSize), tableFile.read(headerWordSize)
            )
            header = self.removeControl(header[0])
            tableFile.seek(headerSize - headerWordSize, 1)
            if header == "":
                break  # get rid of empty lines at the end
            if header in ints:
                value = struct.unpack("i", tableFile.read(4))[0]
                tableFile.seek(dataSize - 4, 1)
            elif header in ints1_8:
                value = struct.unpack("8b", tableFile.read(8))
                tableFile.seek(dataSize - 8, 1)
            elif header in doubles:
                value = struct.unpack("d", tableFile.read(8))[0]
                tableFile.seek(dataSize - 8, 1)
            else:
                value = struct.unpack("{}s".format(dataSize), tableFile.read(dataSize))
                value = self.removeControl(value[0])
            tableData[header] = value
            # increment bytes read
            bytesRead += increment
        tableFile.close()
        return tableData

    def removeControl(self, inBytes: bytes) -> str:
        """Remove control characters from byte strings
        
        Parameters
        ----------
        inBytes : bytes
            Bytes from which to remove control 
        
        Returns
        -------
        str :
            Decodes bytes object with control character removed
        """

        inBytes = inBytes.strip(b"\x00")
        return inBytes.decode()

    def headersFromTable(self, tableData: Dict) -> Tuple[Dict, List]:
        """Populate the headers from the table values
        
        Parameters
        ----------
        tableData : OrederedDict
            Ordered dictionary with table data
        
        Returns
        -------
        headers : Dict
            Dictionary of general headers
        chanHeaders : Dict
            List of channel headers
        """

        # initialise storage
        headers = {}
        chanHeaders = []
        # get the sample freqs for each ts file
        self.tsSampleFreqs = []
        for tsNum in self.tsNums:
            self.tsSampleFreqs.append(tableData["SRL{}".format(tsNum)])
        # for sample frequency, use the continuous channel
        headers["sample_freq"] = self.tsSampleFreqs[self.continuousI]
        # these are the unix time stamps
        firstDate, firstTime, lastDate, lastTime = self.getDates(tableData)
        # the start date is equal to the time of the first record
        headers["start_date"] = firstDate
        headers["start_time"] = firstTime
        datetimeStart = datetime.strptime(
            "{} {}".format(firstDate, firstTime), "%Y-%m-%d %H:%M:%S.%f"
        )
        # the stop date
        datetimeLast = datetime.strptime(
            "{} {}".format(lastDate, lastTime), "%Y-%m-%d %H:%M:%S.%f"
        )
        # records are usually equal to one second (beginning on 0 and ending on the last sample before the next 0)
        datetimeStop = datetimeLast + timedelta(
            seconds=(1.0 - 1.0 / headers["sample_freq"])
        )
        # put the stop date and time in the headers
        headers["stop_date"] = datetimeStop.strftime("%Y-%m-%d")
        headers["stop_time"] = datetimeStop.strftime("%H:%M:%S.%f")
        # here calculate number of samples
        deltaSeconds = (datetimeStop - datetimeStart).total_seconds()
        # calculate number of samples - have to add one because the time given in SPAM recording is the actual time of the last sample
        numSamples = round(deltaSeconds * headers["sample_freq"]) + 1
        headers["num_samples"] = numSamples
        headers["ats_data_file"] = self.continuousF
        # deal with the channel headers
        # now want to do this in the correct order
        # chan headers should reflect the order in the data
        chans = ["Ex", "Ey", "Hx", "Hy", "Hz"]
        chanOrder = []
        for chan in chans:
            chanOrder.append(tableData["CH{}".format(chan.upper())])
        # sort the lists in the right order based on chanOrder
        chanOrder, chans = (
            list(x)
            for x in zip(*sorted(zip(chanOrder, chans), key=lambda pair: pair[0]))
        )
        for chan in chans:
            chanH = self.chanDefaults()
            # set the sample frequency from the main headers
            chanH["sample_freq"] = headers["sample_freq"]
            # channel output information (sensor_type, channel_type, ts_lsb, pos_x1, pos_x2, pos_y1, pos_y2, pos_z1, pos_z2, sensor_sernum)
            chanH["ats_data_file"] = self.dataF[self.continuousI]
            chanH["num_samples"] = numSamples
            # channel information
            chanH["channel_type"] = consistentChans(chan)  # consistent chan naming

            # magnetic channels only
            if isMagnetic(chanH["channel_type"]):
                chanH["sensor_sernum"] = tableData["{}SN".format(chan.upper())][-4:]
                chanH["sensor_type"] = "Phoenix"
                # channel input information (gain_stage1, gain_stage2, hchopper, echopper)
                chanH["gain_stage1"] = tableData["HGN"]
                chanH["gain_stage2"] = 1

            # electric channels only
            if isElectric(chanH["channel_type"]):
                # the distances
                if chan == "Ex":
                    chanH["pos_x1"] = float(tableData["EXLN"]) / 2.0
                    chanH["pos_x2"] = chanH["pos_x1"]
                if chan == "Ey":
                    chanH["pos_y1"] = float(tableData["EYLN"]) / 2.0
                    chanH["pos_y2"] = chanH["pos_y1"]
                # channel input information (gain_stage1, gain_stage2, hchopper, echopper)
                chanH["gain_stage1"] = tableData["EGN"]
                chanH["gain_stage2"] = 1

            # append chanHeaders to the list
            chanHeaders.append(chanH)

        # data information (meas_channels, sample_freq)
        headers["meas_channels"] = len(chans)  # this gets reformatted to an int later
        # return the headers and chanHeaders from this file
        return headers, chanHeaders

    def getDates(self, tableData) -> Tuple[str, str, str, str]:
        """Get recording dates (start and end time)
        
        Parameters
        ----------
        tableData : OrederedDict
            Ordered dictionary with table data
        
        Returns
        -------
        firstDate : str
            Date of first sample as string
        firstTime : str
            Time of first sample as string
        lastDate : str
            Date of last sample as string
        lastTime : str
            Time of last sample as string
        """

        firstSecond = tableData["FTIM"][0]
        firstMinute = tableData["FTIM"][1]
        firstHour = tableData["FTIM"][2]
        firstDay = tableData["FTIM"][3]
        firstMonth = tableData["FTIM"][4]
        firstYear = tableData["FTIM"][5]
        firstCentury = tableData["FTIM"][-1]
        firstDate = "{:02d}{:02d}-{:02d}-{:02d}".format(
            firstCentury, firstYear, firstMonth, firstDay
        )
        firstTime = "{:02d}:{:02d}:{:02d}.000".format(
            firstHour, firstMinute, firstSecond
        )
        # this is the start time of the last record
        lastSecond = tableData["LTIM"][0]
        lastMinute = tableData["LTIM"][1]
        lastHour = tableData["LTIM"][2]
        lastDay = tableData["LTIM"][3]
        lastMonth = tableData["LTIM"][4]
        lastYear = tableData["LTIM"][5]
        lastCentury = tableData["LTIM"][-1]
        lastDate = "{:02d}{:02d}-{:02d}-{:02d}".format(
            lastCentury, lastYear, lastMonth, lastDay
        )
        lastTime = "{:02d}:{:02d}:{:02d}.000".format(lastHour, lastMinute, lastSecond)
        return firstDate, firstTime, lastDate, lastTime

    def checkSamples(self) -> None:
        """Check the number of samples for all the timeseries (ts) files
        
        Recall, the format is 3 bytes two's complement per sample
        """

        self.recordStarts = {}
        self.recordScans = {}
        self.recordBytes = {}
        self.recordSampleStarts = {}
        self.recordSampleStops = {}
        # loop over the tsNums
        samplesDict = {}
        for dFileName in self.dataF:
            ts = int(dFileName[-1])
            self.recordStarts[ts] = []
            self.recordScans[ts] = []
            self.recordBytes[ts] = []
            self.recordSampleStarts[ts] = []
            self.recordSampleStops[ts] = []
            # start number of samples at 0
            samples = 0
            # get file size in samples
            numBytes = os.path.getsize(dFileName)
            bytesread = 0
            # now run through the file and figure out the number of samples
            dFile = open(dFileName, "rb")
            while bytesread < numBytes:
                # read 32 bytes tag
                numScans, numChans, dateString = self.readTag(dFile)
                self.recordBytes[ts].append(bytesread + self.tagByteSize)
                dataBytes = numScans * numChans * self.sampleByteSize
                dFile.seek(dataBytes, 1)
                bytesread += self.tagByteSize + dataBytes
                # save the record start times and scan lengths
                self.recordStarts[ts].append(dateString)
                self.recordScans[ts].append(numScans)
                # save the sample starts
                self.recordSampleStarts[ts].append(samples)
                # increment the number of samples
                # recall, a scan is all channels recorded at one time
                # this is equivalent to one sample
                samples += numScans  # this is the count
                # sample stop is samples -1 because inclusive of the current sample
                self.recordSampleStops[ts].append(samples - 1)
            dFile.close()
            # save number of samples in dict
            samplesDict[ts] = samples
            # logFile.close()

        self.tsNumSamples = []
        for tsNum in self.tsNums:
            self.tsNumSamples.append(samplesDict[tsNum])

        # check the samples of the continuous file
        if self.tsNumSamples[self.continuousI] != self.getNumSamples():
            self.printWarning(
                "Number of samples calculated from times is different to that in file"
            )
            self.printWarning(
                "{} samples in file, {} calculated from time".format(
                    self.tsNumSamples[self.continuousI], self.getNumSamples()
                )
            )

    def reformatHigh(self, path: str, **kwargs) -> None:
        """Write out high frequency time series in internal format
        
        Parameters
        ----------
        path : str
            Directory to write out the reformatted time series
        ts : List[int], optional
            A list of the high frequency ts files to reformat. By default, all of the higher frequency recordings are reformatted
        """

        writer = DataWriterInternal()
        for idx, ts in enumerate(self.tsNums):
            if "ts" in kwargs and ts not in kwargs["ts"]:
                continue  # do not reformat this one
            # let's get the headers
            headers = self.getHeaders()
            chanHeaders, chanMap = self.getChanHeaders()
            chans = self.getChannels()
            # now go through the different ts files to get ready to output
            if ts == self.continuous:
                continue
            sampleFreq = self.tsSampleFreqs[idx]
            # set sample frequency in headers
            headers["sample_freq"] = sampleFreq
            for cH in chanHeaders:
                cH["sample_freq"] = sampleFreq
            # now open the data file
            dFile = open(self.dataF[idx], "rb")
            # each record has to be read separately and then compare time to previous
            outStartTime = datetime.strptime(
                self.recordStarts[ts][0], "%Y-%m-%d %H:%M:%S.%f"
            )
            # set up the data dictionary
            data = {}
            for record, startDate in enumerate(self.recordStarts[ts]):
                # start date is a string
                startByte = self.recordBytes[ts][record]
                startDateTime = datetime.strptime(startDate, "%Y-%m-%d %H:%M:%S.%f")
                # read the record - numpy does not support 24 bit two's complement (3 bytes) - hence use struct
                bytesToRead = (
                    self.recordScans[ts][record]
                    * self.sampleByteSize
                    * self.getNumChannels()
                )
                dFile.seek(startByte, 0)  # seek to start byte from start of file
                dataBytes = dFile.read(bytesToRead)
                dataRead = self.twosComplement(dataBytes)
                dataRecord = {}
                for chan in chans:
                    # as it is the same order as in the header file
                    chanIndex = self.chanMap[chan]
                    dataRecord[chan] = dataRead[
                        chanIndex : self.recordScans[ts][record]
                        * self.getNumChannels() : self.getNumChannels()
                    ]
                # need to compare to previous record
                if record != 0 and startDateTime != prevEndTime:
                    # then need to write out the current data before saving the new data
                    # write out current data
                    outStopTime = prevEndTime - timedelta(
                        seconds=1.0 / sampleFreq
                    )  # because inclusive of first sample (previous end time for continuity comparison)
                    # calculate number of samples
                    numSamples = data[chans[0]].size
                    headers["start_date"] = outStartTime.strftime("%Y-%m-%d")
                    headers["start_time"] = outStartTime.strftime("%H:%M:%S.%f")
                    headers["stop_date"] = outStopTime.strftime("%Y-%m-%d")
                    headers["stop_time"] = outStopTime.strftime("%H:%M:%S.%f")
                    headers["num_samples"] = numSamples
                    for cH in chanHeaders:
                        cH["start_date"] = headers["start_date"]
                        cH["start_time"] = headers["start_time"]
                        cH["stop_date"] = headers["stop_date"]
                        cH["stop_time"] = headers["stop_time"]
                        cH["num_samples"] = numSamples
                    # get the outpath
                    dataOutpath = os.path.join(
                        path,
                        "meas_ts{}_{}_{}".format(
                            ts,
                            outStartTime.strftime("%Y-%m-%d-%H-%M-%S"),
                            outStopTime.strftime("%Y-%m-%d-%H-%M-%S"),
                        ),
                    )
                    # create the timeData object
                    comment = "Unscaled samples for interval {} to {} read in from measurement {}".format(
                        outStartTime, outStopTime, self.dataF[idx]
                    )
                    timeData = TimeData(
                        sampleFreq=self.getSampleFreq(),
                        startTime=outStartTime,
                        stopTime=outStopTime,
                        data=data,
                        comments=comment,
                    )
                    # write out
                    writer.setOutPath(dataOutpath)
                    writer.writeData(headers, chanHeaders, timeData)
                    # then save current data
                    outStartTime = startDateTime
                    data = copy.deepcopy(dataRecord)
                    prevEndTime = startDateTime + timedelta(
                        seconds=((1.0 / sampleFreq) * self.recordScans[ts][record])
                    )
                else:
                    # then record == 0 or startDateTime == prevEndTime
                    # update prevEndTime
                    prevEndTime = startDateTime + timedelta(
                        seconds=((1.0 / sampleFreq) * self.recordScans[ts][record])
                    )
                    if record == 0:
                        data = copy.deepcopy(dataRecord)
                        continue
                    # otherwise, want to concatenate the data
                    for chan in chans:
                        data[chan] = np.concatenate((data[chan], dataRecord[chan]))
            # close the data file
            dFile.close()

    def reformatContinuous(self, path: str):
        """Write out the continuous time series in internal format
        
        Parameters
        ----------
        path : str
            Path to write out reformatted continuous recording
        """

        writer = DataWriterInternal()
        outpath = "meas_ts{}_{}_{}".format(
            self.continuous,
            self.getStartDatetime().strftime("%Y-%m-%d-%H-%M-%S"),
            self.getStopDatetime().strftime("%Y-%m-%d-%H-%M-%S"),
        )
        outpath = os.path.join(path, outpath)
        writer.setOutPath(outpath)
        headers = self.getHeaders()
        chanHeaders, chanMap = self.getChanHeaders()
        writer.writeData(headers, chanHeaders, self.getPhysicalSamples(), physical=True)

    def reformat(self, path):
        """Write out all recorded time series to internal format
        
        Parameters
        ----------
        path : str
            Path to write out reformatted recordings
        """

        self.reformatContinuous(path)
        self.reformatHigh(path)

    def printDataFileList(self) -> List[str]:
        """Information about the data files as a list of strings
        
        Returns
        -------
        List[str]
            List of information about the data files
        """

        textLst = []
        textLst.append("TS File\t\tSampling frequency (Hz)\t\tNum Samples")
        for dF, tsF, tsN in zip(self.dataF, self.tsSampleFreqs, self.tsNumSamples):
            textLst.append("{}\t\t{}\t\t{}".format(os.path.basename(dF), tsF, tsN))
        textLst.append(
            "Continuous data file: {}".format(os.path.basename(self.continuousF))
        )
        return textLst

    def printDataFileInfo(self):
        """Print a list of the data files"""

        blockPrint(
            "{} Data File List".format(self.__class__.__name__),
            self.printDataFileList(),
        )

    def printTableFileList(self) -> List[str]:
        """Information about the table file as a list of strings
        
        Returns
        -------
        List[str]
            List of information about table file content
        """

        textLst = []
        for h, v in list(self.tableData.items()):
            textLst.append("{} = {}".format(h, v))
        return textLst

    def printTableFileInfo(self):
        """Print table file info"""

        blockPrint(
            "{} Table File Info".format(self.__class__.__name__),
            self.printTableFileList(),
        )



from os.path import basename, splitext
import numpy as np
import collections
from typing import List, Dict, IO, Any

# import from package
from resistics.ioHandlers.ioHandler import IOHandler
from resistics.dataObjects.transferFunctionData import TransferFunctionData
from resistics.utilities.utilsPrint import listToString, arrayToString


class TransferFunctionWriter(IOHandler):
    """Class for reading and writing maskData

    Transfer function files are located in:
    project -> transFuncData -> site -> filepath

	Attributes
	----------
    filepath : str 
        Transfer function file
    headers : OrderedDict()
        Transfer function file header information

	Methods
	-------
	__init__(filepath, tfData)
		Initialise the transferFunctionWriter
	defaultHeaders()
        Get a dictionary of default header values
    setHeaders(**kwargs)
        Set header values
    addHeader(header, val)
        Add a new header or update an existing one
    setPathAndData(filepath, tfData)
        Set a new filepath for writing and new transfer function data
    write()
        Write transfer function data to internal format transfer function file
    writeEdi()
        Write transfer function data to Edi file
    writeNumericBlock(lines)
        Write a numeric block for Edi files
    printList()
        Class status returned as list of strings     
	"""

    def __init__(self, filepath: str, tfData: TransferFunctionData, **kwargs) -> None:
        """Initialise

        Parameters
        ----------
        filepath : str
            Path to write a transfer function file file
        tfData : TransferFunctionData
            The transfer function data
        sampleFreq : float, optional
            The sampling frequency of the original time data
        insite : str
            The input site
        inchans : List[str]
            The input channels
        outsite : str
            The output site
        outchans : List[str]
            The output channels
        remotesite : str
            The remote reference site
        remotechans : List[str]
            The remote reference chans used
        """

        self.filepath: str = filepath
        self.tfData: TransferFunctionData = tfData
        self.headers: Dict = self.defaultHeaders()
        for h in self.headers:
            if h in kwargs:
                self.headers[h] = kwargs[h]
        # do the polarisations
        self.polarisations = []
        if len(self.headers["inchans"]) > 0 and len(self.headers["outchans"]) > 0:
            for oC in self.headers["outchans"]:
                for iC in self.headers["inchans"]:
                    self.polarisations.append("{}{}".format(oC, iC))
        else:
            self.polarisations = sorted(tfData.polarisations)

    def defaultHeaders(self) -> Dict[str, Any]:
        """Get the default header values

        Returns
        -------
        default : Dict
            A dictionary with default header keys and values
        """

        default = collections.OrderedDict()
        default["sampleFreq"] = 0
        default["insite"] = False
        default["inchans"] = []
        default["outsite"] = False
        default["outchans"] = []
        default["remotesite"] = False
        default["remotechans"] = []
        return default

    def setHeaders(self, **kwargs):
        """Set header values

        Parameters
        ----------
        sampleFreq : float, optional
            The sampling frequency of the original time data
        insite : str
            The input site
        inchans : List[str]
            The input channels
        outsite : str
            The output site
        outchans : List[str]
            The output channels
        remotesite : str
            The remote reference site
        remotechans : List[str]
            The remote reference chans used
        """        

        for h in self.headers:
            if h in kwargs:
                self.headers[h] = kwargs[h]

    def addHeader(self, header: str, value: Any) -> None:
        """Add a header value for writing out
        
        Parameters
        ----------
        header : str
            The name of the header
        value : Any
            The value for the header
        """

        self.headers[header] = value

    def setPathAndData(self, filepath: str, tfData: TransferFunctionData) -> None:
        """Add a header value for writing out
        
        Parameters
        ----------
        filepath : str
            The filepath to write out to
        tfData : TransferFunctionData
            Transfer function data object
        """

        self.filepath = filepath
        self.tfData = tfData

    def write(self):
        """Write transfer function data in internal format file"""

        outF = open(self.filepath, "w")
        self.printText("Writing out transfer function data to file {}".format(self.filepath))
        # write out the headers
        for h, v in self.headers.items():
            if isinstance(v, list):
                v = listToString(v)
                if len(v) == 0:
                    v = "False"
            outF.write("{} = {}\n".format(h, v))

        # write evaluation frequencies
        outF.write("Evaluation frequencies\n")
        outF.write("{}\n".format(arrayToString(self.tfData.freq)))

        # now need to write out the other data
        for pol in self.polarisations:
            outF.write("Z-{}\n".format(pol))
            outF.write("{}\n".format(arrayToString(self.tfData.data[pol])))

        # variances
        for pol in self.polarisations:
            outF.write("Var-{}\n".format(pol))
            outF.write("{}\n".format(arrayToString(self.tfData.variances[pol])))

        outF.close()

    def writeEdi(self) -> None:
        """Write transfer function data in Edi file"""

        # Write the head and other meta information
        # Then write the values section by section
        outF = open(self.filepath, "w")
        outF.write(">HEAD\n\n")
        # for key, value in self.head.items():
        #     outF.write('{}={}\n'.format(key,value))
        # write out info
        outF.write("\n\n>INFO MAXINFO 5000\n\nempty\n\n")
        # write out define meas
        outF.write(">=DEFINEMEAS\n\nMAXCHAN=7\nMAXRUN=99\nMAXMEAS=9999\n")
        outF.write("UNITS=M\nREFLAT=+47:32:06.08\nREFLONG=+08:01:51.55\nREFELEV=0.00")
        outF.write(">HMEAS ID= 11.001 CHTYPE=HX X= 0 Y= 0 AZM= 0.\n")
        outF.write(">HMEAS ID= 12.001 CHTYPE=HY X= 0 Y= 0 AZM= 90.\n")
        outF.write(">HMEAS ID= 13.001 CHTYPE=HZ X= 0 Y= 0 AZM= 0.\n")
        outF.write(">EMEAS ID= 14.001 CHTYPE=EX X= 0 Y= 0 X2= 0 Y2= 0\n")
        outF.write(">EMEAS ID= 15.001 CHTYPE=EY X= 0 Y= 0 X2= 0 Y2= 0\n")
        outF.write("\n\n")
        # write out MTsect
        outF.write(">=MTSECT\n\nSECTID={}\nNFREQ={:d}\n\n")
        outF.write("HX= 11.001\nHY= 12.001\nHZ= 13.001\nEX= 14.001\nEY= 15.001\n\n")
        # write out frequencies
        numFreq = self.tfData.freq.size
        outF.write(">FREQ NFREQ={:d} ORDER=DEC // {:d}\n".format(numFreq, numFreq))
        self.writeNumericBlock(outF, self.tfData.freq)
        # write out impedances and variances
        for pol in self.polarisations:
            d = "{}{}".format(pol[1], pol[3])
            d = d.upper()
            # need to do a real and imaginary part
            outF.write("\n>Z{}R ROT=0.0 // {:d}\n".format(d, numFreq))
            self.writeNumericBlock(outF, self.tfData.data[pol].real)
            outF.write("\n>Z{}I ROT=0.0 // {:d}\n".format(d, numFreq))
            self.writeNumericBlock(outF, self.tfData.data[pol].imag)
            outF.write("\n>Z{}.VAR ROT=0.0 // {:d}\n".format(d, numFreq))
            self.writeNumericBlock(outF, self.tfData.variances[pol].real)
        # write end of file
        outF.write("\n>END")

    def writeNumericBlock(self, outF: IO[str], vals: np.ndarray) -> None:
        """Write numeric blocks for edi files

        The format is: 
        5 values on a line
        Each value has 5 decimal places and is in scientific notation

        Parameters
        ----------
        outF : file object
            Output file
        vals : np.ndarray
            Numeric files to write out
        """

        numVals = len(vals)
        for i in range(0, numVals):
            if i > 0 and i % 5 == 0:
                outF.write("\n")
            outF.write(" {:.5e}".format(vals[i]))

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : List[str]
            List of strings with information
        """

        textLst: List[str] = []
        textLst.append("Filepath = {}".format(self.filepath))
        textLst.append("Polarisations = {}".format(self.polarisations))
        textLst.append("Headers set")
        for h, val in self.headers.items():
            textLst.append("\t{} = {}".format(h, val))
        return textLst



import numpy as np
import matplotlib.pyplot as plt
from typing import List, Dict, Tuple

# import from package
from resistics.dataObjects.dataObject import DataObject
from resistics.utilities.utilsPlotter import (
    getViewFonts,
    getTransferFunctionFigSize,
    transferFunctionColours,
)
from resistics.utilities.utilsPrint import listToString, arrayToString


class TransferFunctionData(DataObject):
    """Class for holding information time data

    Attributes
    ----------
    freq : np.ndarray
        The evaluation frequencies for which transfer functions were estimated
    period : np.ndarray
        The evaluation periods for which the transfer functions were estimated (1/freq)
    data : Dict
        The transfer function data keyed by polarisations. Data is in the units E = mV, H = nT
    polarisations : List[str]
        List of polarisations in the data
    variances : Dict
        The variances of the estimates to give an idea of uncertainty

    Methods
    -------
    __init__(freq, data, variances)
        Initialise the transfer function data
    getComponent(component)
        Set data with parameters
    getVariance(component)
        Add a comment to the dataset
    getResAndPhase(component) 
        A datetime array of the sample times
    getResAndPhaseErrors(component)
        View the spectra data 
    view()
        View the impedance tensor components
    viewTipper()
        View the tipper
    printList()
        Class status returned as list of strings        

    Notes
    -----
    Information about data units
    Magnetic permeability in nT . m / A
    Electric (E) data is in mV/m 
    Magnetic (H) data is in nT
    Z = E/H is in mV / m . nT
    Units of resistance = Ohm = V / A        
    T = H/H is dimensionless              
    """

    def __init__(self, freq: np.ndarray, data: Dict, variances: Dict):
        """Initialise and set object parameters

        Parameters
        ----------
        freq : np.ndarray
            Evaluation frequency array
        data : Dict
            Dictionary of data keyed by polarisation and with array values. Data units expected to be E = mV, H = nT
        variances : Dict
            Dictionary of uncertainties keyed by polarisation and with array values 
        """

        self.freq: np.ndarray = freq
        self.period: np.ndarray = np.reciprocal(self.freq)
        self.data: Dict = data
        self.polarisations: List[str] = sorted(self.data.keys())
        self.variances: Dict = variances

    def getComponent(self, component: str) -> np.ndarray:
        """Get data for a polarisation

        Parameters
        ----------
        component : str
            The polarisation to return the data for
        
        Returns
        -------
        np.ndarray
            The data for the polarisation component
        """

        return self.data[component]

    def getVariance(self, component: str) -> np.ndarray:
        """Get the variance for a component

        Parameters
        ----------
        component : str
            The polarisation to return the data for
        
        Returns
        -------
        np.ndarray
            The data for the polarisation component
        """

        return self.variances[component]

    def getResAndPhase(self, component: str) -> Tuple[np.ndarray, np.ndarray]:
        """Return the resistivity and phase for a component

        The raw data is expected to be units E = mV, H = nT
        Resisitivity is calculated as 0.2 * period * (abs(data))^2
        Phase is calculated as the complex angle, unwrapped and then shifted for components ExHx and ExHy to put them in 0-90 degree quadrant

        Parameters
        ----------
        component : str
            The polarisation to return the data for
        
        Returns
        -------
        res : np.ndarray
            The resistivity for the evaluation frequencies
        phase : np.ndarray
            The phase for the evaluation frequencies
        """

        data = self.getComponent(component)
        res = 0.2 * self.period * np.power(np.absolute(data), 2)
        phase = np.angle(data)
        # check, can we unwrap into specific quadrant
        phase = np.unwrap(phase)
        # convert to degrees
        phase = phase * 180 / np.pi
        if component == "ExHx" or component == "ExHy":
            # do modulo 360, see if this helps
            phase = np.mod(phase, 360)
            phase = phase - 180
        return res, phase

    def getResAndPhaseErrors(self, component: str) -> Tuple[np.ndarray, np.ndarray]:
        """Return the resistivity and phase errors for a component

        The raw data is expected to be units E = mV, H = nT
        Errors are calculated parameterically (for confidence intervals)
        Error in resistivity is given by 1.96 * sqrt(2 * period * res * var / 5.0)
        Error in phase is given by 1.96 * (180 / pi) * (sqrt(var / 2.0) / abs(data))

        todo: Is there are better way to calculate errors

        Parameters
        ----------
        component : str
            The polarisation to return the data for
        
        Returns
        -------
        resError : np.ndarray
            The errors in resistivity data for the evaluation frequencies
        phaseError : np.ndarray
            The errors in phase data for the evaluation frequencies
        """

        data = self.getComponent(component)
        res, phase = self.getResAndPhase(component)
        var = self.getVariance(component)
        resError = 1.96 * np.sqrt(2 * self.period * res * var / 5.0)
        phaseError = 1.96 * (180 / np.pi) * (np.sqrt(var / 2.0) / np.absolute(data))
        # calculate uncertainties on apparent resistivity and phase
        return resError, phaseError

    def getTipper(self):
        """Return tipper length and angle

        The tipper are the Tx = HzHx and Ty = HzHy components
        The tipper length is sqrt(Re(Tx)^2 + Re(Ty)^2)
        The tipper angle is arctan (Re(Ty)/Re(Tx))

        This needs HzHx and HzHy to be components of the transfer function
        
        Returns
        -------
        tipperLength : np.ndarray
            The tipper length
        tipperAngle : np.ndarray
            The tipper angle
        """

        if "HzHy" not in self.polarisations or "HzHx" not in self.polarisations:
            return False, False
        txRe = np.real(self.getComponent("HzHx"))
        txIm = np.imag(self.getComponent("HzHx"))
        tyRe = np.real(self.getComponent("HzHy"))
        tyIm = np.imag(self.getComponent("HzHy"))
        tipperLength = np.sqrt(np.power(txRe, 2) + np.power(tyRe, 2))
        tipperAngleRe = np.arctan(tyRe / txRe)
        tipperAngleIm = np.arctan(tyIm / txIm)
        return tipperLength, tipperAngleRe, tipperAngleIm

    def viewImpedance(self, **kwargs) -> plt.figure:
        """Plots the transfer function data

        For resistivity data, both axes are log scale (period and resistivity). For phase data, period is in log scale and phase is linear scale.
        Units, x axis is seconds, resistivity is Ohm m and phase is degrees.

        Parameters
        ----------
        polarisations : List[str], optional
            Polarisations to plot
        fig : matplotlib.pyplot.figure, optional
            A figure object
        oneplot : bool, optional   
            Boolean flag for plotting all polarisations on one plot rather than separate plots               
        colours : Dict[str, str], optional
            Colours dictionary for plotting impedance components 
        mk : str, optional
            Plot marker type
        ls : str, optional
            Line style  
        plotfonts : Dict, optional
            A dictionary of plot fonts
        label : str, optional
            Label for the plots
        xlim : List, optional
            Limits for the x axis
        res_ylim : List, optional
            Limits for the resistivity y axis
        phase_ylim : List, optional
            Limits for the phase y axis 

        Returns
        -------
        plt.figure
            Matplotlib figure object            
        """

        polarisations = (
            kwargs["polarisations"] if "polarisations" in kwargs else self.polarisations
        )

        # limits
        xlim = kwargs["xlim"] if "xlim" in kwargs else [1e-3, 1e4]
        res_ylim = kwargs["res_ylim"] if "res_ylim" in kwargs else [1e-2, 1e3]
        phase_ylim = kwargs["phase_ylim"] if "phase_ylim" in kwargs else [-30, 120]

        # markers
        colours = (
            kwargs["colours"] if "colours" in kwargs else transferFunctionColours()
        )
        mk = kwargs["mk"] if "mk" in kwargs else "o"
        ls = kwargs["ls"] if "ls" in kwargs else "none"
        plotfonts = kwargs["plotfonts"] if "plotfonts" in kwargs else getViewFonts()

        # calculate number of rows and columns
        oneplot = False
        if "oneplot" in kwargs and kwargs["oneplot"]:
            oneplot = True
        nrows = 2
        ncols = 1 if oneplot else len(polarisations)
        # a multiplier to make sure all the components end up on the right plot
        plotNumMult = 0 if ncols > 1 else 1

        # plot
        if "fig" in kwargs:
            fig = plt.figure(kwargs["fig"].number)
        else:
            figsize = getTransferFunctionFigSize(oneplot, len(polarisations))
            fig = plt.figure(figsize=figsize)

        st = fig.suptitle(
            "Impedance tensor apparent resistivity and phase",
            fontsize=plotfonts["suptitle"],
        )
        st.set_y(0.98)

        for idx, pol in enumerate(polarisations):
            res, phase = self.getResAndPhase(pol)
            resError, phaseError = self.getResAndPhaseErrors(pol)
            label = kwargs["label"] + " - {}".format(pol) if "label" in kwargs else pol
            # plot resistivity
            ax1 = plt.subplot(nrows, ncols, idx + 1 - plotNumMult * idx)
            # the title
            if not oneplot:
                plt.title("Polarisation {}".format(pol), fontsize=plotfonts["title"])
            else:
                plt.title(
                    "Polarisations {}".format(listToString(polarisations)),
                    fontsize=plotfonts["title"],
                )
            # plot the data
            ax1.errorbar(
                self.period,
                res,
                yerr=resError,
                ls=ls,
                marker=mk,
                markersize=9,
                markerfacecolor="white",
                markeredgecolor=colours[pol],
                mew=1.1,
                color=colours[pol],
                ecolor=colours[pol],
                elinewidth=1.0,
                capsize=4,
                barsabove=False,
                label=label,
            )
            ax1.set_xscale("log")
            ax1.set_yscale("log")
            ax1.set_aspect("equal", adjustable="box")
            # axis options
            plt.ylabel("Apparent Res. [Ohm m]", fontsize=plotfonts["axisLabel"])
            plt.xlim(xlim)
            plt.ylim(res_ylim)
            # set tick sizes
            for lab in ax1.get_xticklabels() + ax1.get_yticklabels():
                lab.set_fontsize(plotfonts["axisTicks"])

            # plot phase
            ax2 = plt.subplot(nrows, ncols, ncols + idx + 1 - plotNumMult * idx)
            # plot the data
            ax2.errorbar(
                self.period,
                phase,
                yerr=phaseError,
                ls="none",
                marker=mk,
                markersize=9,
                markerfacecolor="white",
                markeredgecolor=colours[pol],
                mew=1.1,
                color=colours[pol],
                ecolor=colours[pol],
                elinewidth=1.0,
                capsize=4,
                barsabove=False,
                label=label,
            )
            ax2.set_xscale("log")
            # axis options
            plt.xlabel("Period [s]", fontsize=plotfonts["axisLabel"])
            plt.ylabel("Phase [degrees]", fontsize=plotfonts["axisLabel"])
            plt.xlim(xlim)
            plt.ylim(phase_ylim)
            # set tick sizes
            for lab in ax2.get_xticklabels() + ax2.get_yticklabels():
                lab.set_fontsize(plotfonts["axisTicks"])

        # add the legend
        for idx, pol in enumerate(polarisations):
            ax1 = plt.subplot(nrows, ncols, idx + 1 - plotNumMult * idx)
            leg = plt.legend(loc="lower left", fontsize=plotfonts["legend"])
            leg.get_frame().set_linewidth(0.0)
            leg.get_frame().set_facecolor("w")
            plt.grid(True, ls="--")
            ax2 = plt.subplot(nrows, ncols, ncols + idx + 1 - plotNumMult * idx)
            leg = plt.legend(loc="lower left", fontsize=plotfonts["legend"])
            leg.get_frame().set_linewidth(0.0)
            leg.get_frame().set_facecolor("w")
            plt.grid(True, ls="--")

        # show if the figure is not in keywords
        if "fig" not in kwargs:
            # layout options
            plt.tight_layout()
            fig.subplots_adjust(top=0.92)
            plt.show()

        return fig

    def viewTipper(self, **kwargs) -> plt.figure:
        """Plots tipper data where available

        For length data, both axes are log scale (period and length). For angle data, period is in log scale and phase is linear scale.
        Units, x axis is seconds, length is dimensionless and angle is degrees.

        Parameters
        ----------
        fig : matplotlib.pyplot.figure, optional
            A figure object
        cols : bool, optional
            There are three tipper plots: tipper length, tipper real angle, tipper imaginary angle. These can either be plotted in rows or columns. Set to True to plot in with one row and three columns.
        colours : str, optional
            Colour of plot
        mk : str, optional
            Plot marker type
        ls : str, optional
            Line style  
        plotfonts : Dict, optional
            A dictionary of plot fonts
        label : str, optional
            Label for the plots
        xlim : List, optional
            Limits for the x axis
        length_ylim : List, optional
            Limits for the length y axis
        angle_ylim : List, optional
            Limits for the angle y axis            
        """

        if "HzHx" not in self.polarisations or "HzHy" not in self.polarisations:
            return False

        # rows or columns
        cols = kwargs["cols"] if "cols" in kwargs else True
        if cols:
            nrows = 1
            ncols = 3
        else:
            nrows = 3
            ncols = 1
        # plot options
        mk = kwargs["mk"] if "mk" in kwargs else "o"
        ls = kwargs["ls"] if "ls" in kwargs else "none"
        plotfonts = kwargs["plotfonts"] if "plotfonts" in kwargs else getViewFonts()

        # limits
        xlim = kwargs["xlim"] if "xlim" in kwargs else [1e-3, 1e4]
        length_ylim = kwargs["length_ylim"] if "length_ylim" in kwargs else [1e-2, 1e3]
        angle_ylim = kwargs["angle_ylim"] if "angle_ylim" in kwargs else [-30, 30]

        fig = (
            plt.figure(kwargs["fig"].number)
            if "fig" in kwargs
            else plt.figure(figsize=(16, 7))
        )
        st = fig.suptitle("Tipper")
        st.set_y(0.98)

        # plot
        tipperLength, tipperAngleRe, tipperAngleIm = self.getTipper()
        # lengthError, angleError = self.getTipperErrors()
        label = kwargs["label"] if "label" in kwargs else "tipper"
        # plot resistivity
        ax1 = plt.subplot(nrows, ncols, 1)
        plt.title("Tipper Length")
        ax1.errorbar(
            self.period,
            tipperLength,
            yerr=tipperLength / 50,
            ls=ls,
            marker=mk,
            markerfacecolor="white",
            markersize=9,
            color="red",
            ecolor="red",
            mew=1.1,
            elinewidth=1.0,
            capsize=4,
            barsabove=False,
            label=label,
        )
        ax1.set_xscale("log")
        ax1.set_yscale("log")
        # axis options
        plt.xlabel("Period [s]", fontsize=plotfonts["axisLabel"])
        plt.ylabel("Length [dimensionless]", fontsize=plotfonts["axisLabel"])
        plt.xlim(xlim)
        plt.ylim(length_ylim)
        # set tick sizes and legend
        for lab in ax1.get_xticklabels() + ax1.get_yticklabels():
            lab.set_fontsize(plotfonts["axisTicks"])
        leg = plt.legend(loc="upper right", fontsize=plotfonts["legend"])
        leg.get_frame().set_linewidth(0.0)
        leg.get_frame().set_facecolor("w")
        plt.grid(True, ls="--")

        # plot real tipper angles
        ax2 = plt.subplot(nrows, ncols, 2)
        plt.title("Tipper Angle Real")
        ax2.errorbar(
            self.period,
            tipperAngleRe,
            yerr=tipperAngleRe / 50,
            ls="none",
            marker=mk,
            markerfacecolor="white",
            markersize=9,
            color="red",
            ecolor="red",
            mew=1.1,
            elinewidth=1.0,
            capsize=4,
            barsabove=False,
            label=label,
        )
        ax2.set_xscale("log")
        # axis options
        plt.xlabel("Period [s]", fontsize=plotfonts["axisLabel"])
        plt.ylabel("Angle [degrees]", fontsize=plotfonts["axisLabel"])
        plt.xlim(xlim)
        plt.ylim(angle_ylim)
        # set tick sizes
        for lab in ax2.get_xticklabels() + ax2.get_yticklabels():
            lab.set_fontsize(plotfonts["axisTicks"])
        leg = plt.legend(loc="upper right", fontsize=plotfonts["legend"])
        leg.get_frame().set_linewidth(0.0)
        leg.get_frame().set_facecolor("w")
        plt.grid(True, ls="--")

        # plot imaginary tipper angles
        ax3 = plt.subplot(nrows, ncols, 3)
        plt.title("Tipper Angle Imaginary")
        ax3.errorbar(
            self.period,
            tipperAngleIm,
            yerr=tipperAngleIm / 50,
            ls="none",
            marker=mk,
            markerfacecolor="white",
            markersize=9,
            color="red",
            ecolor="red",
            mew=1.1,
            elinewidth=1.0,
            capsize=4,
            barsabove=False,
            label=label,
        )
        ax3.set_xscale("log")
        # axis options
        plt.xlabel("Period [s]", fontsize=plotfonts["axisLabel"])
        plt.ylabel("Angle [degrees]", fontsize=plotfonts["axisLabel"])
        plt.xlim(xlim)
        plt.ylim(angle_ylim)
        # set tick sizes
        for lab in ax3.get_xticklabels() + ax3.get_yticklabels():
            lab.set_fontsize(plotfonts["axisTicks"])
        leg = plt.legend(loc="upper right", fontsize=plotfonts["legend"])
        leg.get_frame().set_linewidth(0.0)
        leg.get_frame().set_facecolor("w")
        plt.grid(True, ls="--")

        # show if the figure is not in keywords
        if "fig" not in kwargs:
            # layout options
            plt.tight_layout()
            fig.subplots_adjust(top=0.90)
            plt.show()
        
        return fig

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : List[str]
            List of strings with information
        """

        textLst = []
        textLst.append("Frequencies [Hz] = {}".format(arrayToString(self.freq)))
        textLst.append("Periods [s] = {}".format(arrayToString(self.period)))
        textLst.append("Polarisations = {}".format(self.polarisations))
        return textLst



import numpy as np
from typing import Dict, Union

# import from package
from resistics.dataObjects.timeData import TimeData
from resistics.utilities.utilsPrint import errorPrint


def polarityReversal(
    timeData: TimeData, reversal: Dict[str, bool], inplace: bool = True
) -> TimeData:
    """Multiply the data by -1 (polarity reversal)
    
    Parameters
    ----------
    timeData : TimeData
        timeData to normalise
    reversal : Dict[str, bool]
        Keys are channels and values are boolean flags for reversing
    inplace : bool, optional
        Whether to manipulate the data inplace

    Returns
    -------
    TimeData
        Normalised time data
    """

    if not inplace:
        timeData = timeData.copy()
    timeData.data = polarityReversalData(timeData.data, reversal)
    timeData.addComment("Polarity reversal with parameters: {}".format(reversal))
    return timeData


def polarityReversalData(
    data: Dict[str, np.ndarray], reversal: Dict[str, bool]
) -> Dict[str, np.ndarray]:
    """Polarity reverse data or simply multiply by -1
    
    Parameters
    ----------
    data : Dict
        Dictionary with channel as keys and data as values
    reversal : Dict[str, bool]
        Keys are channels and values are boolean flags for reversing
        
    Returns
    -------
    Dict
        Dictionary with channel as keys and normalised data as values
    """

    for c in data:
        if c in reversal and reversal[c]:
            data[c] = data[c] * -1
    return data


def intdiv(nom: Union[int, float], div: Union[int, float]) -> int:
    """Return an integer result of division

    The division is expected to be exact and ensures an integer return rather than float.
    Code execution will exit if division is not exact
    
    Parameters
    ----------
    nom : int, float
        Nominator
    div : int, float
        Divisor    

    Returns
    -------
    out : int
        Result of division
    """

    if nom % div == 0:
        return nom // div
    else:
        errorPrint(
            "utilsMath::intdiv",
            "intdiv assumes exits upon having a remainder to make sure errors are not propagated through the code",
            quitRun=True,
        )
        return 0

